from fastapi import APIRouter, Depends, HTTPException, Request, Response
from pydantic import BaseModel
from sqlalchemy import func, nullslast, or_, orm, select, update
from sqlalchemy.orm import contains_eager, lazyload, sessionmaker

from ultima_scraper_db.databases.ultima_archive.api.client import (
    UAClient,
    get_ua_client,
)
from ultima_scraper_db.databases.ultima_archive.schemas.templates.site import (
    SubscriptionModel,
    UserInfoModel,
    UserModel,
)

router = APIRouter(
    prefix="/sites",
    tags=["sites"],
    responses={404: {"description": "Not found"}},
)


@router.get("/")
async def get_sites(
    ua_client: UAClient = Depends(get_ua_client),
):
    database_api = ua_client.database_api

    async with database_api.create_management_api() as management_api:
        sites = await management_api.get_sites()
        return sites



from typing import TYPE_CHECKING, Any

from sqlalchemy import BigInteger, ForeignKey, MetaData, inspect
from sqlalchemy.dialects.postgresql.base import PGCompiler
from sqlalchemy.ext.asyncio import AsyncAttrs
from sqlalchemy.ext.compiler import compiles  # type: ignore
from sqlalchemy.ext.declarative import declared_attr
from sqlalchemy.orm import DeclarativeBase, Mapped, mapped_column, relationship
from sqlalchemy.sql import expression

if TYPE_CHECKING:
    from ultima_scraper_db.databases.ultima_archive.schemas.templates.site import (
        MassMessageModel,
        MessageModel,
        PostModel,
        StoryModel,
    )

    content_model_types = StoryModel | PostModel | MessageModel | MassMessageModel


class CustomFuncs:
    class utcnow(expression.FunctionElement):  # type: ignore
        from sqlalchemy.types import DateTime

        type = DateTime()
        inherit_cache = True

    @compiles(utcnow, "postgresql")
    def pg_utcnow(element, compiler: PGCompiler, **kw: Any):
        return "TIMEZONE('utc', CURRENT_TIMESTAMP)"

    @compiles(utcnow, "mssql")
    def ms_utcnow(element, compiler: PGCompiler, **kw: Any):
        return "GETUTCDATE()"


class DefaultContentTypes(AsyncAttrs):
    __mapper_args__ = {"eager_defaults": True}
    story_id: Mapped[int] = mapped_column(
        BigInteger, ForeignKey("x_stories.id"), nullable=True
    )
    post_id: Mapped[int] = mapped_column(
        BigInteger, ForeignKey("x_posts.id"), nullable=True
    )
    message_id: Mapped[int] = mapped_column(
        BigInteger, ForeignKey("x_messages.id"), nullable=True
    )
    mass_message_id: Mapped[int] = mapped_column(
        BigInteger, ForeignKey("x_mass_messages.id"), nullable=True
    )
    media_id: Mapped[int] = mapped_column(BigInteger, ForeignKey("media.id"))

    @declared_attr
    def story(self) -> Mapped["StoryModel"]:
        return relationship("StoryModel")

    @declared_attr
    def post(self) -> Mapped["PostModel"]:
        return relationship("PostModel")

    @declared_attr
    def message(self) -> Mapped["MessageModel"]:
        return relationship("MessageModel")

    @declared_attr
    def mass_message(self) -> Mapped["MassMessageModel"]:
        return relationship("MassMessageModel")

    async def get_content(self):
        from ultima_scraper_db.databases.ultima_archive.schemas.templates.site import (
            ContentMediaAssoModel,
        )

        exclusions = ["media"]

        await self.awaitable_attrs.story
        await self.awaitable_attrs.post
        await self.awaitable_attrs.message
        await self.awaitable_attrs.mass_message
        for key, _ in inspect(ContentMediaAssoModel).relationships.items():
            if any(x for x in exclusions if x == key):
                continue
            final_value: content_model_types = getattr(self, key)
            if final_value:
                return final_value
        raise Exception("Content not set")

    async def get_content_media_asso(self):
        from ultima_scraper_db.databases.ultima_archive.schemas.templates.site import (
            ContentMediaAssoModel,
        )

        exclusions = ["media"]

        for key, _ in inspect(ContentMediaAssoModel).relationships.items():
            if any(x for x in exclusions if x == key):
                continue
            final_key = f"{key}_id"
            final_value: int = getattr(self, final_key)
            if final_value:
                return final_key, final_value
        raise Exception("Content not set")

    def set_content(self, content: "content_model_types"):
        match type(content).__name__:
            case "StoryModel":
                self.story = content
            case "PostModel":
                self.post = content
            case "MessageModel":
                self.message = content
            case "MassMessageModel":
                self.mass_message = content
            case _:
                raise Exception("Content type not found")


class UltimaBase(AsyncAttrs, DeclarativeBase):
    __mapper_args__ = {"eager_defaults": True}


class ManagementTemplate(UltimaBase):
    __abstract__ = True
    __table_args__ = {"schema": "management"}


class SiteTemplate(UltimaBase):
    __abstract__ = True


from ultima_scraper_api import SUPPORTED_SITES

# Used for alembic upgrades
from ultima_scraper_db.databases.ultima_archive.schemas import management as _
from ultima_scraper_db.databases.ultima_archive.schemas.templates import site as _


def create_merged_metadata(base: UltimaBase | Any, remove_this: dict[str, Any] = {}):
    site_metadatas: list[MetaData] = []
    for supp_site in SUPPORTED_SITES:
        site_metadatas.append(MetaData(schema=supp_site.lower()))
    merged_metadata = MetaData()
    for _k, t in base.metadata.tables.items():
        if not t.schema:
            for site_metadata in site_metadatas:

                def referred_schema_fn(
                    table: Any,
                    to_schema: Any,
                    constraint: Any,
                    referred_schema: str | None,
                ):
                    if referred_schema is not None:
                        return referred_schema
                    return site_metadata.schema

                new_table = t.to_metadata(site_metadata)
                new_table.to_metadata(
                    merged_metadata, referred_schema_fn=referred_schema_fn
                )
                # if "status" in t.c:
                #     t.c.status.schema = site_metadata.schema
                # if _k == "jobs":
                #     pass
                pass
        else:
            if remove_this:
                if t.key == remove_this["table"]:
                    my_col_instance = [
                        col for col in t._columns if col.name == remove_this["column"]
                    ][0]
                    t._columns.remove(my_col_instance)
            t.to_metadata(merged_metadata)
    return merged_metadata


merged_metadata = create_merged_metadata(UltimaBase)



from pathlib import Path

from fastapi import APIRouter, Depends, HTTPException, Request, Response
from pydantic import BaseModel
from sqlalchemy import or_, orm, select, update
from sqlalchemy.orm import contains_eager, lazyload, sessionmaker

from ultima_scraper_db.databases.ultima_archive.api.client import UAClient
from ultima_scraper_db.databases.ultima_archive.schemas.templates.site import (
    FilePathModel,
)

router = APIRouter(
    prefix="/filepaths",
    tags=["filepaths"],
    responses={404: {"description": "Not found"}},
)


@router.get("/")
async def get_filepath(request: Request, site_name: str, media_id: int):
    database_api = UAClient.database_api
    async with database_api.create_site_api(site_name) as site_api:
        filepaths = await site_api.get_filepaths(media_id=media_id)
        return filepaths


@router.get("/{site_name}/{content_type}/{content_id}")
async def get_filepath_by_content_id(
    request: Request, site_name: str, content_type: str, content_id: int
):
    database_api = UAClient.database_api
    async with database_api.create_site_api(site_name) as site_api:
        filepaths = await site_api.get_filepaths(post_id=content_id)
        for filepath in filepaths:
            if not filepath.downloaded and Path(filepath.filepath).exists():
                filepath.downloaded = True
        return filepaths



from fastapi import APIRouter

routers: list[APIRouter] = []
router_names = [
    "users",
    "jobs",
    "client",
    "remote_urls",
    "notifications",
    "filepaths",
    "media_detections",
    "sites",
    "hosts",
    "search",
]

for name in router_names:
    module_path = f"ultima_scraper_db.databases.ultima_archive.api.routers.{name}"
    router = getattr(__import__(module_path, fromlist=["router"]), "router")
    routers.append(router)



from ultima_scraper_api import SUPPORTED_SITES
from ultima_scraper_collection.config import UltimaScraperCollectionConfig

from ultima_scraper_db.databases.ultima_archive.api.client import UAClient
from ultima_scraper_db.databases.ultima_archive.filters import AuthedInfoFilter
from ultima_scraper_db.databases.ultima_archive.management_api import ManagementAPI
from ultima_scraper_db.databases.ultima_archive.site_api import SiteAPI
from ultima_scraper_db.managers.database_manager import Database, DatabaseAPI_, Schema


class ArchiveAPI(DatabaseAPI_):
    def __init__(self, database: Database) -> None:
        from ultima_scraper_collection.managers.server_manager import ServerManager

        super().__init__(database)

        self.management_api: ManagementAPI = self.create_management_api()
        self.management_schema: Schema = self.database.schemas["management"]
        self.site_apis: dict[str, SiteAPI] = {}
        for supported_site in SUPPORTED_SITES:
            supported_site = supported_site.lower()
            self.site_apis[supported_site] = self.create_site_api(supported_site)
        self.server_manager = ServerManager(self)

    async def init(self):

        await self.server_manager.init()
        return self

    def create_management_api(self):
        return ManagementAPI(self.database.schemas["management"])

    def create_site_api(self, name: str):
        return SiteAPI(self.database.schemas[name.lower()])

    def get_site_api(self, name: str):
        return self.site_apis[name.lower()]

    def find_site_api(self, name: str):
        return self.site_apis[name.lower()]

    async def get_sites(self):
        async with self.create_management_api() as management_api:
            sites = await management_api.get_sites()
            return sites

    async def activate_fast_api(
        self,
        database: Database,
        config: "UltimaScraperCollectionConfig",
        port: int = 2140,
    ):
        fast_api = UAClient(database_api=ArchiveAPI(database))
        await fast_api.init(config)
        self.activate_api(fast_api, port)
        self.fast_api = fast_api
        return self.fast_api

    async def update_authed_users(self):
        import ultima_scraper_api
        from ultima_scraper_api.helpers.main_helper import get_date_range_past_days

        from ultima_scraper_db.databases.ultima_archive.schemas.templates.site import (
            UserModel,
        )

        for site_name in SUPPORTED_SITES:
            site_api = ultima_scraper_api.select_api(site_name)
            async with self.create_site_api(site_name) as site_db_api:
                authed_info_filter = AuthedInfoFilter(
                    exclude_between_dates=get_date_range_past_days(), active=True
                )
                db_users = await site_db_api.get_users(
                    authed_info_filter=authed_info_filter,
                    order_by=UserModel.last_checked_at.asc(),
                )
                for db_user in db_users:
                    for db_auth_info in db_user.find_auths():
                        auth_details = db_auth_info.convert_to_auth_details(site_name)
                        async with site_api.login_context(
                            auth_details.export()
                        ) as authed:
                            if not authed:
                                print(f"User {db_user.username} failed to login.")
                                await db_auth_info.deactivate()
                            else:
                                print(f"User {db_user.username} logged in.")
                                await site_db_api.create_or_update_user(
                                    authed.user, db_user, performer_optimize=True
                                )
                                break
                    await site_db_api.get_session().commit()
        return True



from datetime import datetime


class AuthedInfoFilter:
    def __init__(
        self,
        between_dates: tuple[datetime, datetime] | None = None,
        exclude_between_dates: tuple[datetime, datetime] | None = None,
        active: bool | None = None,
    ) -> None:
        self.between_dates = between_dates
        self.exclude_between_dates = exclude_between_dates
        self.active = active



from datetime import datetime

from fastapi import APIRouter, Depends, HTTPException, Request, Response
from pydantic import BaseModel

from ultima_scraper_db.databases.ultima_archive.api.client import UAClient

router = APIRouter(
    prefix="/remote_urls",
    tags=["remote_urls"],
    responses={404: {"description": "Not found"}},
)


class RemoteURLData(BaseModel):
    user_id: int
    host_id: int
    url: str
    part: int = 0
    uploaded_at: float


class Performer(BaseModel):
    identifier: int | str
    site_name: str


@router.get("/{site_name}/{user_id}")
async def get_remote_urls(request: Request, site_name: str, user_id: int):
    database_api = UAClient.database_api

    site_api = database_api.create_site_api(site_name)
    async with site_api as site_api:
        remote_urls = await site_api.get_remote_urls(user_id)
        return remote_urls


@router.patch("/{site_name}")
async def update_remote_url(
    request: Request,
    site_name: str,
    remote_url: RemoteURLData,
):
    database_api = UAClient.database_api

    site_api = database_api.create_site_api(site_name)
    async with site_api as site_api:
        remote_url_db = await site_api.create_or_update_remote_url(
            remote_url.user_id,
            host_id=remote_url.host_id,
            url=remote_url.url,
            part=remote_url.part,
            uploaded_at=datetime.fromtimestamp(remote_url.uploaded_at),
        )
        return remote_url_db


@router.post("/bulk")
async def get_remote_urls_bulk(request: Request, performers: list[Performer]):
    database_api = UAClient.database_api

    grouped_performers = {}
    for performer in performers:
        site_name = performer.site_name
        if site_name not in grouped_performers:
            grouped_performers[site_name] = []
        grouped_performers[site_name].append(performer)

    all_remote_urls = {}
    for site_name, site_performers in grouped_performers.items():
        performer_ids = [int(performer.identifier) for performer in site_performers]
        site_api = database_api.create_site_api(site_name)
        async with site_api as site_api:
            remote_urls = await site_api.get_bulk_remote_urls(performer_ids)
            if site_name not in all_remote_urls:
                all_remote_urls[site_name] = {}
            for performer in site_performers:
                performer_urls = [
                    url
                    for url in remote_urls
                    if url.user_id == int(performer.identifier)
                ]
                all_remote_urls[site_name][performer.identifier] = performer_urls
    return all_remote_urls



"""empty message

Revision ID: ac840f4e3061
Revises: 
Create Date: 2023-12-30 16:44:56.996566

"""

import sqlalchemy as sa
from alembic import op

# revision identifiers, used by Alembic.
revision = "ac840f4e3061"
down_revision = None
branch_labels = None
depends_on = None


def upgrade() -> None:
    # ### commands auto generated by Alembic - please adjust! ###
    op.create_table(
        "users",
        sa.Column("id", sa.BigInteger(), autoincrement=False, nullable=False),
        sa.Column("username", sa.Text(), nullable=True),
        sa.Column("balance", sa.Float(), server_default="0", nullable=False),
        sa.Column("spend", sa.Boolean(), server_default="false", nullable=False),
        sa.Column("performer", sa.Boolean(), server_default="false", nullable=False),
        sa.Column("favorite", sa.Boolean(), server_default="false", nullable=False),
        sa.Column("active", sa.Boolean(), server_default="true", nullable=False),
        sa.Column("downloaded_at", sa.TIMESTAMP(timezone=True), nullable=True),
        sa.Column("last_checked_at", sa.TIMESTAMP(timezone=True), nullable=True),
        sa.Column("join_date", sa.TIMESTAMP(timezone=True), nullable=True),
        sa.Column(
            "created_at",
            sa.TIMESTAMP(timezone=True),
            server_default=sa.text("TIMEZONE('utc', CURRENT_TIMESTAMP)"),
            nullable=False,
        ),
        sa.Column("updated_at", sa.TIMESTAMP(timezone=True), nullable=True),
        sa.PrimaryKeyConstraint("id"),
        schema="fansly",
    )
    op.create_table(
        "hosts",
        sa.Column("id", sa.Integer(), nullable=False),
        sa.Column("name", sa.Text(), nullable=False),
        sa.Column("identifier", sa.Text(), nullable=True),
        sa.Column("password", sa.Text(), nullable=True),
        sa.Column("source", sa.Boolean(), server_default="True", nullable=False),
        sa.Column("active", sa.Boolean(), server_default="True", nullable=False),
        sa.PrimaryKeyConstraint("id"),
        schema="management",
    )
    op.create_table(
        "servers",
        sa.Column("id", sa.Integer(), nullable=False),
        sa.Column("name", sa.Text(), nullable=False),
        sa.Column("ip", sa.Text(), nullable=False),
        sa.Column("mac_address", sa.Text(), nullable=False),
        sa.Column("job_limit", sa.SmallInteger(), server_default="10", nullable=False),
        sa.Column("active", sa.SmallInteger(), server_default="1", nullable=False),
        sa.PrimaryKeyConstraint("id"),
        schema="management",
    )
    op.create_table(
        "sites",
        sa.Column("id", sa.Integer(), nullable=False),
        sa.Column("name", sa.Text(), nullable=False),
        sa.Column("db_name", sa.Text(), nullable=False),
        sa.Column("url", sa.Text(), nullable=False),
        sa.Column("active", sa.Boolean(), server_default="true", nullable=False),
        sa.Column("size", sa.BigInteger(), server_default="0", nullable=False),
        sa.Column("user_id_checkpoint", sa.BigInteger(), nullable=True),
        sa.PrimaryKeyConstraint("id"),
        schema="management",
    )
    op.create_table(
        "users",
        sa.Column("id", sa.BigInteger(), autoincrement=False, nullable=False),
        sa.Column("username", sa.Text(), nullable=True),
        sa.Column("balance", sa.Float(), server_default="0", nullable=False),
        sa.Column("spend", sa.Boolean(), server_default="false", nullable=False),
        sa.Column("performer", sa.Boolean(), server_default="false", nullable=False),
        sa.Column("favorite", sa.Boolean(), server_default="false", nullable=False),
        sa.Column("active", sa.Boolean(), server_default="true", nullable=False),
        sa.Column("downloaded_at", sa.TIMESTAMP(timezone=True), nullable=True),
        sa.Column("last_checked_at", sa.TIMESTAMP(timezone=True), nullable=True),
        sa.Column("join_date", sa.TIMESTAMP(timezone=True), nullable=True),
        sa.Column(
            "created_at",
            sa.TIMESTAMP(timezone=True),
            server_default=sa.text("TIMEZONE('utc', CURRENT_TIMESTAMP)"),
            nullable=False,
        ),
        sa.Column("updated_at", sa.TIMESTAMP(timezone=True), nullable=True),
        sa.PrimaryKeyConstraint("id"),
        schema="onlyfans",
    )
    op.create_table(
        "bought_content",
        sa.Column("id", sa.Integer(), autoincrement=True, nullable=False),
        sa.Column("supplier_id", sa.BigInteger(), nullable=False),
        sa.Column("buyer_id", sa.BigInteger(), nullable=False),
        sa.ForeignKeyConstraint(
            ["buyer_id"],
            ["fansly.users.id"],
        ),
        sa.ForeignKeyConstraint(
            ["supplier_id"],
            ["fansly.users.id"],
        ),
        sa.PrimaryKeyConstraint("id"),
        schema="fansly",
    )
    op.create_table(
        "favorite_users",
        sa.Column("id", sa.Integer(), nullable=False),
        sa.Column("user_id", sa.BigInteger(), nullable=False),
        sa.Column("notification", sa.Boolean(), server_default="true", nullable=False),
        sa.ForeignKeyConstraint(
            ["user_id"],
            ["fansly.users.id"],
        ),
        sa.PrimaryKeyConstraint("id"),
        sa.UniqueConstraint("user_id"),
        schema="fansly",
    )
    op.create_table(
        "histo_user_infos",
        sa.Column("id", sa.Integer(), nullable=False),
        sa.Column("user_id", sa.BigInteger(), nullable=False),
        sa.Column("name", sa.Text(), nullable=True),
        sa.Column("price", sa.Float(), server_default="0", nullable=False),
        sa.Column("description", sa.Text(), nullable=True),
        sa.Column("promotion", sa.Boolean(), server_default="true", nullable=False),
        sa.Column("post_count", sa.Integer(), server_default="0", nullable=False),
        sa.Column("media_count", sa.Integer(), server_default="0", nullable=False),
        sa.Column("image_count", sa.Integer(), server_default="0", nullable=False),
        sa.Column("video_count", sa.Integer(), server_default="0", nullable=False),
        sa.Column("audio_count", sa.Integer(), server_default="0", nullable=False),
        sa.Column("stream_count", sa.Integer(), server_default="0", nullable=False),
        sa.Column(
            "archived_post_count", sa.Integer(), server_default="0", nullable=False
        ),
        sa.Column(
            "private_archived_post_count",
            sa.Integer(),
            server_default="0",
            nullable=False,
        ),
        sa.Column("favourited_count", sa.Integer(), server_default="0", nullable=False),
        sa.Column("favourites_count", sa.Integer(), server_default="0", nullable=False),
        sa.Column(
            "subscribers_count", sa.Integer(), server_default="0", nullable=False
        ),
        sa.Column("size", sa.BigInteger(), server_default="0", nullable=False),
        sa.Column("location", sa.Text(), nullable=True),
        sa.Column("website", sa.Text(), nullable=True),
        sa.Column(
            "created_at",
            sa.TIMESTAMP(timezone=True),
            server_default=sa.text("now()"),
            nullable=False,
        ),
        sa.ForeignKeyConstraint(
            ["user_id"],
            ["fansly.users.id"],
        ),
        sa.PrimaryKeyConstraint("id"),
        schema="fansly",
    )
    op.create_table(
        "jobs",
        sa.Column("id", sa.Integer(), nullable=False),
        sa.Column("site_id", sa.BigInteger(), nullable=False),
        sa.Column("user_id", sa.BigInteger(), nullable=False),
        sa.Column("user_username", sa.Text(), nullable=False),
        sa.Column("category", sa.Text(), nullable=False),
        sa.Column("server_id", sa.BigInteger(), nullable=False),
        sa.Column("host_id", sa.BigInteger(), nullable=True),
        sa.Column("skippable", sa.Boolean(), server_default="false", nullable=False),
        sa.Column("priority", sa.Boolean(), server_default="false", nullable=False),
        sa.Column("active", sa.Boolean(), server_default="true", nullable=False),
        sa.Column("completed_at", sa.TIMESTAMP(timezone=True), nullable=True),
        sa.ForeignKeyConstraint(
            ["host_id"],
            ["management.hosts.id"],
        ),
        sa.ForeignKeyConstraint(
            ["server_id"],
            ["management.servers.id"],
        ),
        sa.ForeignKeyConstraint(
            ["site_id"],
            ["management.sites.id"],
        ),
        sa.ForeignKeyConstraint(
            ["user_id"],
            ["fansly.users.id"],
        ),
        sa.PrimaryKeyConstraint("id"),
        sa.UniqueConstraint("site_id", "user_id", "category"),
        schema="fansly",
    )
    op.create_table(
        "mass_message_stats",
        sa.Column("id", sa.BigInteger(), autoincrement=False, nullable=False),
        sa.Column("user_id", sa.BigInteger(), nullable=False),
        sa.Column("media_count", sa.SmallInteger(), nullable=False),
        sa.Column("buyer_count", sa.SmallInteger(), nullable=False),
        sa.Column("sent_count", sa.Integer(), nullable=False),
        sa.Column("view_count", sa.Integer(), nullable=False),
        sa.ForeignKeyConstraint(
            ["user_id"],
            ["fansly.users.id"],
        ),
        sa.PrimaryKeyConstraint("id"),
        schema="fansly",
    )
    op.create_table(
        "media",
        sa.Column("id", sa.BigInteger(), autoincrement=False, nullable=False),
        sa.Column("user_id", sa.BigInteger(), nullable=False),
        sa.Column("url", sa.Text(), nullable=True),
        sa.Column("size", sa.BigInteger(), server_default="0", nullable=False),
        sa.Column("category", sa.String(), nullable=True),
        sa.Column("preview", sa.Boolean(), server_default="false", nullable=False),
        sa.Column("created_at", sa.TIMESTAMP(timezone=True), nullable=True),
        sa.ForeignKeyConstraint(
            ["user_id"],
            ["fansly.users.id"],
        ),
        sa.PrimaryKeyConstraint("id"),
        schema="fansly",
    )
    op.create_table(
        "notifications",
        sa.Column("id", sa.Integer(), nullable=False),
        sa.Column("user_id", sa.BigInteger(), nullable=True),
        sa.Column("category", sa.String(), nullable=False),
        sa.Column("sent_discord", sa.Boolean(), server_default="false", nullable=False),
        sa.Column(
            "sent_telegram", sa.Boolean(), server_default="false", nullable=False
        ),
        sa.Column(
            "created_at",
            sa.TIMESTAMP(timezone=True),
            server_default=sa.text("TIMEZONE('utc', CURRENT_TIMESTAMP)"),
            nullable=False,
        ),
        sa.ForeignKeyConstraint(
            ["user_id"],
            ["fansly.users.id"],
        ),
        sa.PrimaryKeyConstraint("id"),
        sa.UniqueConstraint("user_id", "category"),
        schema="fansly",
    )
    op.create_table(
        "remote_urls",
        sa.Column("id", sa.Integer(), nullable=False),
        sa.Column("user_id", sa.BigInteger(), nullable=False),
        sa.Column("host_id", sa.BigInteger(), nullable=False),
        sa.Column("url", sa.Text(), nullable=True),
        sa.Column("exists", sa.Boolean(), server_default="False", nullable=False),
        sa.Column("uploaded_at", sa.TIMESTAMP(timezone=True), nullable=False),
        sa.Column("downloaded_at", sa.TIMESTAMP(timezone=True), nullable=True),
        sa.ForeignKeyConstraint(
            ["host_id"],
            ["management.hosts.id"],
        ),
        sa.ForeignKeyConstraint(
            ["user_id"],
            ["fansly.users.id"],
        ),
        sa.PrimaryKeyConstraint("id"),
        schema="fansly",
    )
    op.create_table(
        "socials",
        sa.Column("id", sa.Integer(), nullable=False),
        sa.Column("user_id", sa.BigInteger(), nullable=False),
        sa.Column("site", sa.String(), nullable=False),
        sa.Column("identifier", sa.String(), nullable=False),
        sa.ForeignKeyConstraint(
            ["user_id"],
            ["fansly.users.id"],
        ),
        sa.PrimaryKeyConstraint("id"),
        schema="fansly",
    )
    op.create_table(
        "subscriptions",
        sa.Column("id", sa.Integer(), nullable=False),
        sa.Column("user_id", sa.BigInteger(), nullable=False),
        sa.Column("subscriber_id", sa.BigInteger(), nullable=False),
        sa.Column("paid_content", sa.Boolean(), server_default="0", nullable=False),
        sa.Column("active", sa.Boolean(), server_default="true", nullable=False),
        sa.Column("downloaded_at", sa.TIMESTAMP(timezone=True), nullable=True),
        sa.Column("created_at", sa.TIMESTAMP(timezone=True), nullable=False),
        sa.Column("renewed_at", sa.TIMESTAMP(timezone=True), nullable=True),
        sa.Column("expires_at", sa.TIMESTAMP(timezone=True), nullable=False),
        sa.ForeignKeyConstraint(
            ["subscriber_id"],
            ["fansly.users.id"],
        ),
        sa.ForeignKeyConstraint(
            ["user_id"],
            ["fansly.users.id"],
        ),
        sa.PrimaryKeyConstraint("id"),
        sa.UniqueConstraint("user_id", "subscriber_id"),
        schema="fansly",
    )
    op.create_table(
        "user_aliases",
        sa.Column("id", sa.Integer(), nullable=False),
        sa.Column("user_id", sa.BigInteger(), nullable=False),
        sa.Column("username", sa.String(length=50), nullable=False),
        sa.ForeignKeyConstraint(
            ["user_id"],
            ["fansly.users.id"],
        ),
        sa.PrimaryKeyConstraint("id"),
        sa.UniqueConstraint("user_id", "username"),
        schema="fansly",
    )
    op.create_table(
        "user_auths",
        sa.Column("id", sa.Integer(), nullable=False),
        sa.Column("user_id", sa.BigInteger(), nullable=False),
        sa.Column("cookie", sa.Text(), nullable=True),
        sa.Column("x_bc", sa.Text(), nullable=True),
        sa.Column("authorization", sa.Text(), nullable=True),
        sa.Column("user_agent", sa.Text(), nullable=False),
        sa.Column("email", sa.Text(), nullable=True),
        sa.Column("password", sa.Text(), nullable=True),
        sa.Column("active", sa.Boolean(), server_default="true", nullable=False),
        sa.ForeignKeyConstraint(
            ["user_id"],
            ["fansly.users.id"],
        ),
        sa.PrimaryKeyConstraint("id"),
        sa.UniqueConstraint("cookie", "authorization"),
        schema="fansly",
    )
    op.create_table(
        "user_infos",
        sa.Column("id", sa.Integer(), nullable=False),
        sa.Column("user_id", sa.BigInteger(), nullable=False),
        sa.Column("name", sa.Text(), nullable=True),
        sa.Column("price", sa.Float(), server_default="0", nullable=False),
        sa.Column("description", sa.Text(), nullable=True),
        sa.Column("promotion", sa.Boolean(), server_default="true", nullable=False),
        sa.Column("sex", sa.SmallInteger(), nullable=True),
        sa.Column("post_count", sa.Integer(), server_default="0", nullable=False),
        sa.Column("media_count", sa.Integer(), server_default="0", nullable=False),
        sa.Column("image_count", sa.Integer(), server_default="0", nullable=False),
        sa.Column("video_count", sa.Integer(), server_default="0", nullable=False),
        sa.Column("audio_count", sa.Integer(), server_default="0", nullable=False),
        sa.Column("stream_count", sa.Integer(), server_default="0", nullable=False),
        sa.Column(
            "archived_post_count", sa.Integer(), server_default="0", nullable=False
        ),
        sa.Column(
            "private_archived_post_count",
            sa.Integer(),
            server_default="0",
            nullable=False,
        ),
        sa.Column("favourited_count", sa.Integer(), server_default="0", nullable=False),
        sa.Column("favourites_count", sa.Integer(), server_default="0", nullable=False),
        sa.Column(
            "subscribers_count", sa.Integer(), server_default="0", nullable=False
        ),
        sa.Column("size", sa.BigInteger(), server_default="0", nullable=False),
        sa.Column("location", sa.Text(), nullable=True),
        sa.Column("website", sa.Text(), nullable=True),
        sa.Column("downloaded_at", sa.TIMESTAMP(timezone=True), nullable=True),
        sa.Column("first_downloaded_at", sa.TIMESTAMP(timezone=True), nullable=True),
        sa.Column("uploaded_at", sa.TIMESTAMP(timezone=True), nullable=True),
        sa.ForeignKeyConstraint(
            ["user_id"],
            ["fansly.users.id"],
        ),
        sa.PrimaryKeyConstraint("id"),
        sa.UniqueConstraint("user_id"),
        schema="fansly",
    )
    op.create_index(
        op.f("ix_fansly_user_infos_sex"),
        "user_infos",
        ["sex"],
        unique=False,
        schema="fansly",
    )
    op.create_table(
        "x_posts",
        sa.Column("id", sa.BigInteger(), autoincrement=False, nullable=False),
        sa.Column("user_id", sa.BigInteger(), nullable=False),
        sa.Column("text", sa.Text(), nullable=True),
        sa.Column("price", sa.Float(), server_default="0", nullable=False),
        sa.Column("deleted", sa.Boolean(), server_default="false", nullable=False),
        sa.Column("archived", sa.Boolean(), server_default="false", nullable=False),
        sa.Column("paid", sa.Boolean(), nullable=True),
        sa.Column("created_at", sa.TIMESTAMP(timezone=True), nullable=False),
        sa.ForeignKeyConstraint(
            ["user_id"],
            ["fansly.users.id"],
        ),
        sa.PrimaryKeyConstraint("id"),
        schema="fansly",
    )
    op.create_table(
        "x_stories",
        sa.Column("id", sa.BigInteger(), autoincrement=False, nullable=False),
        sa.Column("user_id", sa.BigInteger(), nullable=False),
        sa.Column("created_at", sa.TIMESTAMP(timezone=True), nullable=False),
        sa.ForeignKeyConstraint(
            ["user_id"],
            ["fansly.users.id"],
        ),
        sa.PrimaryKeyConstraint("id"),
        schema="fansly",
    )
    op.create_table(
        "bought_content",
        sa.Column("id", sa.Integer(), autoincrement=True, nullable=False),
        sa.Column("supplier_id", sa.BigInteger(), nullable=False),
        sa.Column("buyer_id", sa.BigInteger(), nullable=False),
        sa.ForeignKeyConstraint(
            ["buyer_id"],
            ["onlyfans.users.id"],
        ),
        sa.ForeignKeyConstraint(
            ["supplier_id"],
            ["onlyfans.users.id"],
        ),
        sa.PrimaryKeyConstraint("id"),
        schema="onlyfans",
    )
    op.create_table(
        "favorite_users",
        sa.Column("id", sa.Integer(), nullable=False),
        sa.Column("user_id", sa.BigInteger(), nullable=False),
        sa.Column("notification", sa.Boolean(), server_default="true", nullable=False),
        sa.ForeignKeyConstraint(
            ["user_id"],
            ["onlyfans.users.id"],
        ),
        sa.PrimaryKeyConstraint("id"),
        sa.UniqueConstraint("user_id"),
        schema="onlyfans",
    )
    op.create_table(
        "histo_user_infos",
        sa.Column("id", sa.Integer(), nullable=False),
        sa.Column("user_id", sa.BigInteger(), nullable=False),
        sa.Column("name", sa.Text(), nullable=True),
        sa.Column("price", sa.Float(), server_default="0", nullable=False),
        sa.Column("description", sa.Text(), nullable=True),
        sa.Column("promotion", sa.Boolean(), server_default="true", nullable=False),
        sa.Column("post_count", sa.Integer(), server_default="0", nullable=False),
        sa.Column("media_count", sa.Integer(), server_default="0", nullable=False),
        sa.Column("image_count", sa.Integer(), server_default="0", nullable=False),
        sa.Column("video_count", sa.Integer(), server_default="0", nullable=False),
        sa.Column("audio_count", sa.Integer(), server_default="0", nullable=False),
        sa.Column("stream_count", sa.Integer(), server_default="0", nullable=False),
        sa.Column(
            "archived_post_count", sa.Integer(), server_default="0", nullable=False
        ),
        sa.Column(
            "private_archived_post_count",
            sa.Integer(),
            server_default="0",
            nullable=False,
        ),
        sa.Column("favourited_count", sa.Integer(), server_default="0", nullable=False),
        sa.Column("favourites_count", sa.Integer(), server_default="0", nullable=False),
        sa.Column(
            "subscribers_count", sa.Integer(), server_default="0", nullable=False
        ),
        sa.Column("size", sa.BigInteger(), server_default="0", nullable=False),
        sa.Column("location", sa.Text(), nullable=True),
        sa.Column("website", sa.Text(), nullable=True),
        sa.Column(
            "created_at",
            sa.TIMESTAMP(timezone=True),
            server_default=sa.text("now()"),
            nullable=False,
        ),
        sa.ForeignKeyConstraint(
            ["user_id"],
            ["onlyfans.users.id"],
        ),
        sa.PrimaryKeyConstraint("id"),
        schema="onlyfans",
    )
    op.create_table(
        "jobs",
        sa.Column("id", sa.Integer(), nullable=False),
        sa.Column("site_id", sa.BigInteger(), nullable=False),
        sa.Column("user_id", sa.BigInteger(), nullable=False),
        sa.Column("user_username", sa.Text(), nullable=False),
        sa.Column("category", sa.Text(), nullable=False),
        sa.Column("server_id", sa.BigInteger(), nullable=False),
        sa.Column("host_id", sa.BigInteger(), nullable=True),
        sa.Column("skippable", sa.Boolean(), server_default="false", nullable=False),
        sa.Column("priority", sa.Boolean(), server_default="false", nullable=False),
        sa.Column("active", sa.Boolean(), server_default="true", nullable=False),
        sa.Column("completed_at", sa.TIMESTAMP(timezone=True), nullable=True),
        sa.ForeignKeyConstraint(
            ["host_id"],
            ["management.hosts.id"],
        ),
        sa.ForeignKeyConstraint(
            ["server_id"],
            ["management.servers.id"],
        ),
        sa.ForeignKeyConstraint(
            ["site_id"],
            ["management.sites.id"],
        ),
        sa.ForeignKeyConstraint(
            ["user_id"],
            ["onlyfans.users.id"],
        ),
        sa.PrimaryKeyConstraint("id"),
        sa.UniqueConstraint("site_id", "user_id", "category"),
        schema="onlyfans",
    )
    op.create_table(
        "mass_message_stats",
        sa.Column("id", sa.BigInteger(), autoincrement=False, nullable=False),
        sa.Column("user_id", sa.BigInteger(), nullable=False),
        sa.Column("media_count", sa.SmallInteger(), nullable=False),
        sa.Column("buyer_count", sa.SmallInteger(), nullable=False),
        sa.Column("sent_count", sa.Integer(), nullable=False),
        sa.Column("view_count", sa.Integer(), nullable=False),
        sa.ForeignKeyConstraint(
            ["user_id"],
            ["onlyfans.users.id"],
        ),
        sa.PrimaryKeyConstraint("id"),
        schema="onlyfans",
    )
    op.create_table(
        "media",
        sa.Column("id", sa.BigInteger(), autoincrement=False, nullable=False),
        sa.Column("user_id", sa.BigInteger(), nullable=False),
        sa.Column("url", sa.Text(), nullable=True),
        sa.Column("size", sa.BigInteger(), server_default="0", nullable=False),
        sa.Column("category", sa.String(), nullable=True),
        sa.Column("preview", sa.Boolean(), server_default="false", nullable=False),
        sa.Column("created_at", sa.TIMESTAMP(timezone=True), nullable=True),
        sa.ForeignKeyConstraint(
            ["user_id"],
            ["onlyfans.users.id"],
        ),
        sa.PrimaryKeyConstraint("id"),
        schema="onlyfans",
    )
    op.create_table(
        "notifications",
        sa.Column("id", sa.Integer(), nullable=False),
        sa.Column("user_id", sa.BigInteger(), nullable=True),
        sa.Column("category", sa.String(), nullable=False),
        sa.Column("sent_discord", sa.Boolean(), server_default="false", nullable=False),
        sa.Column(
            "sent_telegram", sa.Boolean(), server_default="false", nullable=False
        ),
        sa.Column(
            "created_at",
            sa.TIMESTAMP(timezone=True),
            server_default=sa.text("TIMEZONE('utc', CURRENT_TIMESTAMP)"),
            nullable=False,
        ),
        sa.ForeignKeyConstraint(
            ["user_id"],
            ["onlyfans.users.id"],
        ),
        sa.PrimaryKeyConstraint("id"),
        sa.UniqueConstraint("user_id", "category"),
        schema="onlyfans",
    )
    op.create_table(
        "remote_urls",
        sa.Column("id", sa.Integer(), nullable=False),
        sa.Column("user_id", sa.BigInteger(), nullable=False),
        sa.Column("host_id", sa.BigInteger(), nullable=False),
        sa.Column("url", sa.Text(), nullable=True),
        sa.Column("exists", sa.Boolean(), server_default="False", nullable=False),
        sa.Column("uploaded_at", sa.TIMESTAMP(timezone=True), nullable=False),
        sa.Column("downloaded_at", sa.TIMESTAMP(timezone=True), nullable=True),
        sa.ForeignKeyConstraint(
            ["host_id"],
            ["management.hosts.id"],
        ),
        sa.ForeignKeyConstraint(
            ["user_id"],
            ["onlyfans.users.id"],
        ),
        sa.PrimaryKeyConstraint("id"),
        schema="onlyfans",
    )
    op.create_table(
        "socials",
        sa.Column("id", sa.Integer(), nullable=False),
        sa.Column("user_id", sa.BigInteger(), nullable=False),
        sa.Column("site", sa.String(), nullable=False),
        sa.Column("identifier", sa.String(), nullable=False),
        sa.ForeignKeyConstraint(
            ["user_id"],
            ["onlyfans.users.id"],
        ),
        sa.PrimaryKeyConstraint("id"),
        schema="onlyfans",
    )
    op.create_table(
        "subscriptions",
        sa.Column("id", sa.Integer(), nullable=False),
        sa.Column("user_id", sa.BigInteger(), nullable=False),
        sa.Column("subscriber_id", sa.BigInteger(), nullable=False),
        sa.Column("paid_content", sa.Boolean(), server_default="0", nullable=False),
        sa.Column("active", sa.Boolean(), server_default="true", nullable=False),
        sa.Column("downloaded_at", sa.TIMESTAMP(timezone=True), nullable=True),
        sa.Column("created_at", sa.TIMESTAMP(timezone=True), nullable=False),
        sa.Column("renewed_at", sa.TIMESTAMP(timezone=True), nullable=True),
        sa.Column("expires_at", sa.TIMESTAMP(timezone=True), nullable=False),
        sa.ForeignKeyConstraint(
            ["subscriber_id"],
            ["onlyfans.users.id"],
        ),
        sa.ForeignKeyConstraint(
            ["user_id"],
            ["onlyfans.users.id"],
        ),
        sa.PrimaryKeyConstraint("id"),
        sa.UniqueConstraint("user_id", "subscriber_id"),
        schema="onlyfans",
    )
    op.create_table(
        "user_aliases",
        sa.Column("id", sa.Integer(), nullable=False),
        sa.Column("user_id", sa.BigInteger(), nullable=False),
        sa.Column("username", sa.String(length=50), nullable=False),
        sa.ForeignKeyConstraint(
            ["user_id"],
            ["onlyfans.users.id"],
        ),
        sa.PrimaryKeyConstraint("id"),
        sa.UniqueConstraint("user_id", "username"),
        schema="onlyfans",
    )
    op.create_table(
        "user_auths",
        sa.Column("id", sa.Integer(), nullable=False),
        sa.Column("user_id", sa.BigInteger(), nullable=False),
        sa.Column("cookie", sa.Text(), nullable=True),
        sa.Column("x_bc", sa.Text(), nullable=True),
        sa.Column("authorization", sa.Text(), nullable=True),
        sa.Column("user_agent", sa.Text(), nullable=False),
        sa.Column("email", sa.Text(), nullable=True),
        sa.Column("password", sa.Text(), nullable=True),
        sa.Column("active", sa.Boolean(), server_default="true", nullable=False),
        sa.ForeignKeyConstraint(
            ["user_id"],
            ["onlyfans.users.id"],
        ),
        sa.PrimaryKeyConstraint("id"),
        sa.UniqueConstraint("cookie", "authorization"),
        schema="onlyfans",
    )
    op.create_table(
        "user_infos",
        sa.Column("id", sa.Integer(), nullable=False),
        sa.Column("user_id", sa.BigInteger(), nullable=False),
        sa.Column("name", sa.Text(), nullable=True),
        sa.Column("price", sa.Float(), server_default="0", nullable=False),
        sa.Column("description", sa.Text(), nullable=True),
        sa.Column("promotion", sa.Boolean(), server_default="true", nullable=False),
        sa.Column("sex", sa.SmallInteger(), nullable=True),
        sa.Column("post_count", sa.Integer(), server_default="0", nullable=False),
        sa.Column("media_count", sa.Integer(), server_default="0", nullable=False),
        sa.Column("image_count", sa.Integer(), server_default="0", nullable=False),
        sa.Column("video_count", sa.Integer(), server_default="0", nullable=False),
        sa.Column("audio_count", sa.Integer(), server_default="0", nullable=False),
        sa.Column("stream_count", sa.Integer(), server_default="0", nullable=False),
        sa.Column(
            "archived_post_count", sa.Integer(), server_default="0", nullable=False
        ),
        sa.Column(
            "private_archived_post_count",
            sa.Integer(),
            server_default="0",
            nullable=False,
        ),
        sa.Column("favourited_count", sa.Integer(), server_default="0", nullable=False),
        sa.Column("favourites_count", sa.Integer(), server_default="0", nullable=False),
        sa.Column(
            "subscribers_count", sa.Integer(), server_default="0", nullable=False
        ),
        sa.Column("size", sa.BigInteger(), server_default="0", nullable=False),
        sa.Column("location", sa.Text(), nullable=True),
        sa.Column("website", sa.Text(), nullable=True),
        sa.Column("downloaded_at", sa.TIMESTAMP(timezone=True), nullable=True),
        sa.Column("first_downloaded_at", sa.TIMESTAMP(timezone=True), nullable=True),
        sa.Column("uploaded_at", sa.TIMESTAMP(timezone=True), nullable=True),
        sa.ForeignKeyConstraint(
            ["user_id"],
            ["onlyfans.users.id"],
        ),
        sa.PrimaryKeyConstraint("id"),
        sa.UniqueConstraint("user_id"),
        schema="onlyfans",
    )
    op.create_index(
        op.f("ix_onlyfans_user_infos_sex"),
        "user_infos",
        ["sex"],
        unique=False,
        schema="onlyfans",
    )
    op.create_table(
        "x_posts",
        sa.Column("id", sa.BigInteger(), autoincrement=False, nullable=False),
        sa.Column("user_id", sa.BigInteger(), nullable=False),
        sa.Column("text", sa.Text(), nullable=True),
        sa.Column("price", sa.Float(), server_default="0", nullable=False),
        sa.Column("deleted", sa.Boolean(), server_default="false", nullable=False),
        sa.Column("archived", sa.Boolean(), server_default="false", nullable=False),
        sa.Column("paid", sa.Boolean(), nullable=True),
        sa.Column("created_at", sa.TIMESTAMP(timezone=True), nullable=False),
        sa.ForeignKeyConstraint(
            ["user_id"],
            ["onlyfans.users.id"],
        ),
        sa.PrimaryKeyConstraint("id"),
        schema="onlyfans",
    )
    op.create_table(
        "x_stories",
        sa.Column("id", sa.BigInteger(), autoincrement=False, nullable=False),
        sa.Column("user_id", sa.BigInteger(), nullable=False),
        sa.Column("created_at", sa.TIMESTAMP(timezone=True), nullable=False),
        sa.ForeignKeyConstraint(
            ["user_id"],
            ["onlyfans.users.id"],
        ),
        sa.PrimaryKeyConstraint("id"),
        schema="onlyfans",
    )
    op.create_table(
        "comments",
        sa.Column("id", sa.BigInteger(), autoincrement=False, nullable=False),
        sa.Column("post_id", sa.BigInteger(), nullable=False),
        sa.Column("reply_id", sa.BigInteger(), nullable=True),
        sa.Column("user_id", sa.BigInteger(), nullable=False),
        sa.Column("giphy_id", sa.String(), nullable=True),
        sa.Column("text", sa.String(), nullable=True),
        sa.Column("likes_count", sa.Integer(), server_default="0", nullable=False),
        sa.Column("created_at", sa.TIMESTAMP(timezone=True), nullable=False),
        sa.ForeignKeyConstraint(
            ["post_id"],
            ["fansly.x_posts.id"],
        ),
        sa.ForeignKeyConstraint(
            ["reply_id"],
            ["fansly.comments.id"],
        ),
        sa.PrimaryKeyConstraint("id"),
        schema="fansly",
    )
    op.create_table(
        "linked_posts",
        sa.Column("id", sa.Integer(), nullable=False),
        sa.Column("post_id", sa.BigInteger(), nullable=False),
        sa.Column("linked_post_id", sa.BigInteger(), nullable=False),
        sa.ForeignKeyConstraint(
            ["linked_post_id"],
            ["fansly.x_posts.id"],
        ),
        sa.ForeignKeyConstraint(
            ["post_id"],
            ["fansly.x_posts.id"],
        ),
        sa.PrimaryKeyConstraint("id"),
        schema="fansly",
    )
    op.create_table(
        "linked_users",
        sa.Column("id", sa.Integer(), nullable=False),
        sa.Column("post_id", sa.BigInteger(), nullable=False),
        sa.Column("linked_user_id", sa.BigInteger(), nullable=False),
        sa.ForeignKeyConstraint(
            ["linked_user_id"],
            ["fansly.users.id"],
        ),
        sa.ForeignKeyConstraint(
            ["post_id"],
            ["fansly.x_posts.id"],
        ),
        sa.PrimaryKeyConstraint("id"),
        schema="fansly",
    )
    op.create_table(
        "media_detections",
        sa.Column("id", sa.Integer(), nullable=False),
        sa.Column("media_id", sa.BigInteger(), nullable=False),
        sa.Column("label", sa.Text(), nullable=False),
        sa.Column("face_female", sa.Boolean(), server_default="false", nullable=False),
        sa.Column(
            "female_genitalia_covered",
            sa.Boolean(),
            server_default="false",
            nullable=False,
        ),
        sa.Column(
            "female_genitalia_exposed",
            sa.Boolean(),
            server_default="false",
            nullable=False,
        ),
        sa.Column(
            "female_breast_covered",
            sa.Boolean(),
            server_default="false",
            nullable=False,
        ),
        sa.Column(
            "female_breast_exposed",
            sa.Boolean(),
            server_default="false",
            nullable=False,
        ),
        sa.Column("face_male", sa.Boolean(), server_default="false", nullable=False),
        sa.Column(
            "male_genitalia_exposed",
            sa.Boolean(),
            server_default="false",
            nullable=False,
        ),
        sa.Column(
            "male_breast_exposed", sa.Boolean(), server_default="false", nullable=False
        ),
        sa.Column(
            "buttocks_covered", sa.Boolean(), server_default="false", nullable=False
        ),
        sa.Column(
            "buttocks_exposed", sa.Boolean(), server_default="false", nullable=False
        ),
        sa.Column("anus_covered", sa.Boolean(), server_default="false", nullable=False),
        sa.Column("anus_exposed", sa.Boolean(), server_default="false", nullable=False),
        sa.Column("feet_covered", sa.Boolean(), server_default="false", nullable=False),
        sa.Column("feet_exposed", sa.Boolean(), server_default="false", nullable=False),
        sa.Column(
            "belly_covered", sa.Boolean(), server_default="false", nullable=False
        ),
        sa.Column(
            "belly_exposed", sa.Boolean(), server_default="false", nullable=False
        ),
        sa.Column(
            "armpits_covered", sa.Boolean(), server_default="false", nullable=False
        ),
        sa.Column(
            "armpits_exposed", sa.Boolean(), server_default="false", nullable=False
        ),
        sa.Column("score", sa.Float(), server_default="0", nullable=False),
        sa.Column("x_coord", sa.Integer(), nullable=True),
        sa.Column("y_coord", sa.Integer(), nullable=True),
        sa.Column("width", sa.Integer(), nullable=True),
        sa.Column("height", sa.Integer(), nullable=True),
        sa.Column(
            "created_at",
            sa.TIMESTAMP(timezone=True),
            server_default=sa.text("TIMEZONE('utc', CURRENT_TIMESTAMP)"),
            nullable=False,
        ),
        sa.ForeignKeyConstraint(
            ["media_id"],
            ["fansly.media.id"],
        ),
        sa.PrimaryKeyConstraint("id"),
        schema="fansly",
    )
    op.create_table(
        "x_mass_messages",
        sa.Column("id", sa.BigInteger(), autoincrement=False, nullable=False),
        sa.Column("user_id", sa.BigInteger(), nullable=False),
        sa.Column("statistic_id", sa.BigInteger(), nullable=True),
        sa.Column("text", sa.Text(), nullable=True),
        sa.Column("price", sa.Float(), nullable=False),
        sa.Column("expires_at", sa.TIMESTAMP(timezone=True), nullable=False),
        sa.Column("created_at", sa.TIMESTAMP(timezone=True), nullable=False),
        sa.ForeignKeyConstraint(
            ["statistic_id"],
            ["fansly.mass_message_stats.id"],
        ),
        sa.ForeignKeyConstraint(
            ["user_id"],
            ["fansly.users.id"],
        ),
        sa.PrimaryKeyConstraint("id"),
        schema="fansly",
    )
    op.create_table(
        "comments",
        sa.Column("id", sa.BigInteger(), autoincrement=False, nullable=False),
        sa.Column("post_id", sa.BigInteger(), nullable=False),
        sa.Column("reply_id", sa.BigInteger(), nullable=True),
        sa.Column("user_id", sa.BigInteger(), nullable=False),
        sa.Column("giphy_id", sa.String(), nullable=True),
        sa.Column("text", sa.String(), nullable=True),
        sa.Column("likes_count", sa.Integer(), server_default="0", nullable=False),
        sa.Column("created_at", sa.TIMESTAMP(timezone=True), nullable=False),
        sa.ForeignKeyConstraint(
            ["post_id"],
            ["onlyfans.x_posts.id"],
        ),
        sa.ForeignKeyConstraint(
            ["reply_id"],
            ["onlyfans.comments.id"],
        ),
        sa.PrimaryKeyConstraint("id"),
        schema="onlyfans",
    )
    op.create_table(
        "linked_posts",
        sa.Column("id", sa.Integer(), nullable=False),
        sa.Column("post_id", sa.BigInteger(), nullable=False),
        sa.Column("linked_post_id", sa.BigInteger(), nullable=False),
        sa.ForeignKeyConstraint(
            ["linked_post_id"],
            ["onlyfans.x_posts.id"],
        ),
        sa.ForeignKeyConstraint(
            ["post_id"],
            ["onlyfans.x_posts.id"],
        ),
        sa.PrimaryKeyConstraint("id"),
        schema="onlyfans",
    )
    op.create_table(
        "linked_users",
        sa.Column("id", sa.Integer(), nullable=False),
        sa.Column("post_id", sa.BigInteger(), nullable=False),
        sa.Column("linked_user_id", sa.BigInteger(), nullable=False),
        sa.ForeignKeyConstraint(
            ["linked_user_id"],
            ["onlyfans.users.id"],
        ),
        sa.ForeignKeyConstraint(
            ["post_id"],
            ["onlyfans.x_posts.id"],
        ),
        sa.PrimaryKeyConstraint("id"),
        schema="onlyfans",
    )
    op.create_table(
        "media_detections",
        sa.Column("id", sa.Integer(), nullable=False),
        sa.Column("media_id", sa.BigInteger(), nullable=False),
        sa.Column("label", sa.Text(), nullable=False),
        sa.Column("face_female", sa.Boolean(), server_default="false", nullable=False),
        sa.Column(
            "female_genitalia_covered",
            sa.Boolean(),
            server_default="false",
            nullable=False,
        ),
        sa.Column(
            "female_genitalia_exposed",
            sa.Boolean(),
            server_default="false",
            nullable=False,
        ),
        sa.Column(
            "female_breast_covered",
            sa.Boolean(),
            server_default="false",
            nullable=False,
        ),
        sa.Column(
            "female_breast_exposed",
            sa.Boolean(),
            server_default="false",
            nullable=False,
        ),
        sa.Column("face_male", sa.Boolean(), server_default="false", nullable=False),
        sa.Column(
            "male_genitalia_exposed",
            sa.Boolean(),
            server_default="false",
            nullable=False,
        ),
        sa.Column(
            "male_breast_exposed", sa.Boolean(), server_default="false", nullable=False
        ),
        sa.Column(
            "buttocks_covered", sa.Boolean(), server_default="false", nullable=False
        ),
        sa.Column(
            "buttocks_exposed", sa.Boolean(), server_default="false", nullable=False
        ),
        sa.Column("anus_covered", sa.Boolean(), server_default="false", nullable=False),
        sa.Column("anus_exposed", sa.Boolean(), server_default="false", nullable=False),
        sa.Column("feet_covered", sa.Boolean(), server_default="false", nullable=False),
        sa.Column("feet_exposed", sa.Boolean(), server_default="false", nullable=False),
        sa.Column(
            "belly_covered", sa.Boolean(), server_default="false", nullable=False
        ),
        sa.Column(
            "belly_exposed", sa.Boolean(), server_default="false", nullable=False
        ),
        sa.Column(
            "armpits_covered", sa.Boolean(), server_default="false", nullable=False
        ),
        sa.Column(
            "armpits_exposed", sa.Boolean(), server_default="false", nullable=False
        ),
        sa.Column("score", sa.Float(), server_default="0", nullable=False),
        sa.Column("x_coord", sa.Integer(), nullable=True),
        sa.Column("y_coord", sa.Integer(), nullable=True),
        sa.Column("width", sa.Integer(), nullable=True),
        sa.Column("height", sa.Integer(), nullable=True),
        sa.Column(
            "created_at",
            sa.TIMESTAMP(timezone=True),
            server_default=sa.text("TIMEZONE('utc', CURRENT_TIMESTAMP)"),
            nullable=False,
        ),
        sa.ForeignKeyConstraint(
            ["media_id"],
            ["onlyfans.media.id"],
        ),
        sa.PrimaryKeyConstraint("id"),
        schema="onlyfans",
    )
    op.create_table(
        "x_mass_messages",
        sa.Column("id", sa.BigInteger(), autoincrement=False, nullable=False),
        sa.Column("user_id", sa.BigInteger(), nullable=False),
        sa.Column("statistic_id", sa.BigInteger(), nullable=True),
        sa.Column("text", sa.Text(), nullable=True),
        sa.Column("price", sa.Float(), nullable=False),
        sa.Column("expires_at", sa.TIMESTAMP(timezone=True), nullable=False),
        sa.Column("created_at", sa.TIMESTAMP(timezone=True), nullable=False),
        sa.ForeignKeyConstraint(
            ["statistic_id"],
            ["onlyfans.mass_message_stats.id"],
        ),
        sa.ForeignKeyConstraint(
            ["user_id"],
            ["onlyfans.users.id"],
        ),
        sa.PrimaryKeyConstraint("id"),
        schema="onlyfans",
    )
    op.create_table(
        "x_messages",
        sa.Column("id", sa.BigInteger(), autoincrement=False, nullable=False),
        sa.Column("user_id", sa.BigInteger(), nullable=False),
        sa.Column("receiver_id", sa.BigInteger(), nullable=True),
        sa.Column("queue_id", sa.BigInteger(), nullable=True),
        sa.Column("text", sa.Text(), nullable=True),
        sa.Column("price", sa.Float(), server_default="0", nullable=False),
        sa.Column("deleted", sa.Boolean(), server_default="false", nullable=False),
        sa.Column("paid", sa.Boolean(), nullable=True),
        sa.Column("verified", sa.Boolean(), server_default="false", nullable=False),
        sa.Column("created_at", sa.TIMESTAMP(timezone=True), nullable=False),
        sa.ForeignKeyConstraint(
            ["queue_id"],
            ["fansly.x_mass_messages.id"],
        ),
        sa.ForeignKeyConstraint(
            ["receiver_id"],
            ["fansly.users.id"],
        ),
        sa.ForeignKeyConstraint(
            ["user_id"],
            ["fansly.users.id"],
        ),
        sa.PrimaryKeyConstraint("id"),
        schema="fansly",
    )
    op.create_table(
        "x_messages",
        sa.Column("id", sa.BigInteger(), autoincrement=False, nullable=False),
        sa.Column("user_id", sa.BigInteger(), nullable=False),
        sa.Column("receiver_id", sa.BigInteger(), nullable=True),
        sa.Column("queue_id", sa.BigInteger(), nullable=True),
        sa.Column("text", sa.Text(), nullable=True),
        sa.Column("price", sa.Float(), server_default="0", nullable=False),
        sa.Column("deleted", sa.Boolean(), server_default="false", nullable=False),
        sa.Column("paid", sa.Boolean(), nullable=True),
        sa.Column("verified", sa.Boolean(), server_default="false", nullable=False),
        sa.Column("created_at", sa.TIMESTAMP(timezone=True), nullable=False),
        sa.ForeignKeyConstraint(
            ["queue_id"],
            ["onlyfans.x_mass_messages.id"],
        ),
        sa.ForeignKeyConstraint(
            ["receiver_id"],
            ["onlyfans.users.id"],
        ),
        sa.ForeignKeyConstraint(
            ["user_id"],
            ["onlyfans.users.id"],
        ),
        sa.PrimaryKeyConstraint("id"),
        schema="onlyfans",
    )
    op.create_table(
        "content_media_asso",
        sa.Column("id", sa.BigInteger(), nullable=False),
        sa.Column("story_id", sa.BigInteger(), nullable=True),
        sa.Column("post_id", sa.BigInteger(), nullable=True),
        sa.Column("message_id", sa.BigInteger(), nullable=True),
        sa.Column("mass_message_id", sa.BigInteger(), nullable=True),
        sa.Column("media_id", sa.BigInteger(), nullable=False),
        sa.ForeignKeyConstraint(
            ["mass_message_id"],
            ["fansly.x_mass_messages.id"],
        ),
        sa.ForeignKeyConstraint(
            ["media_id"],
            ["fansly.media.id"],
        ),
        sa.ForeignKeyConstraint(
            ["message_id"],
            ["fansly.x_messages.id"],
        ),
        sa.ForeignKeyConstraint(
            ["post_id"],
            ["fansly.x_posts.id"],
        ),
        sa.ForeignKeyConstraint(
            ["story_id"],
            ["fansly.x_stories.id"],
        ),
        sa.PrimaryKeyConstraint("id"),
        sa.UniqueConstraint("mass_message_id", "media_id"),
        sa.UniqueConstraint("message_id", "media_id"),
        sa.UniqueConstraint("post_id", "media_id"),
        sa.UniqueConstraint("story_id", "media_id"),
        schema="fansly",
    )
    op.create_table(
        "filepaths",
        sa.Column("id", sa.Integer(), autoincrement=True, nullable=False),
        sa.Column("filepath", sa.String(length=255), nullable=False),
        sa.Column("preview", sa.Boolean(), server_default="false", nullable=False),
        sa.Column("downloaded", sa.Boolean(), server_default="false", nullable=False),
        sa.Column("story_id", sa.BigInteger(), nullable=True),
        sa.Column("post_id", sa.BigInteger(), nullable=True),
        sa.Column("message_id", sa.BigInteger(), nullable=True),
        sa.Column("mass_message_id", sa.BigInteger(), nullable=True),
        sa.Column("media_id", sa.BigInteger(), nullable=False),
        sa.ForeignKeyConstraint(
            ["mass_message_id"],
            ["fansly.x_mass_messages.id"],
        ),
        sa.ForeignKeyConstraint(
            ["media_id"],
            ["fansly.media.id"],
        ),
        sa.ForeignKeyConstraint(
            ["message_id"],
            ["fansly.x_messages.id"],
        ),
        sa.ForeignKeyConstraint(
            ["post_id"],
            ["fansly.x_posts.id"],
        ),
        sa.ForeignKeyConstraint(
            ["story_id"],
            ["fansly.x_stories.id"],
        ),
        sa.PrimaryKeyConstraint("id"),
        sa.UniqueConstraint("mass_message_id", "media_id"),
        sa.UniqueConstraint("message_id", "media_id"),
        sa.UniqueConstraint("post_id", "media_id"),
        sa.UniqueConstraint("story_id", "media_id"),
        schema="fansly",
    )
    op.create_table(
        "content_media_asso",
        sa.Column("id", sa.BigInteger(), nullable=False),
        sa.Column("story_id", sa.BigInteger(), nullable=True),
        sa.Column("post_id", sa.BigInteger(), nullable=True),
        sa.Column("message_id", sa.BigInteger(), nullable=True),
        sa.Column("mass_message_id", sa.BigInteger(), nullable=True),
        sa.Column("media_id", sa.BigInteger(), nullable=False),
        sa.ForeignKeyConstraint(
            ["mass_message_id"],
            ["onlyfans.x_mass_messages.id"],
        ),
        sa.ForeignKeyConstraint(
            ["media_id"],
            ["onlyfans.media.id"],
        ),
        sa.ForeignKeyConstraint(
            ["message_id"],
            ["onlyfans.x_messages.id"],
        ),
        sa.ForeignKeyConstraint(
            ["post_id"],
            ["onlyfans.x_posts.id"],
        ),
        sa.ForeignKeyConstraint(
            ["story_id"],
            ["onlyfans.x_stories.id"],
        ),
        sa.PrimaryKeyConstraint("id"),
        sa.UniqueConstraint("mass_message_id", "media_id"),
        sa.UniqueConstraint("message_id", "media_id"),
        sa.UniqueConstraint("post_id", "media_id"),
        sa.UniqueConstraint("story_id", "media_id"),
        schema="onlyfans",
    )
    op.create_table(
        "filepaths",
        sa.Column("id", sa.Integer(), autoincrement=True, nullable=False),
        sa.Column("filepath", sa.String(length=255), nullable=False),
        sa.Column("preview", sa.Boolean(), server_default="false", nullable=False),
        sa.Column("downloaded", sa.Boolean(), server_default="false", nullable=False),
        sa.Column("story_id", sa.BigInteger(), nullable=True),
        sa.Column("post_id", sa.BigInteger(), nullable=True),
        sa.Column("message_id", sa.BigInteger(), nullable=True),
        sa.Column("mass_message_id", sa.BigInteger(), nullable=True),
        sa.Column("media_id", sa.BigInteger(), nullable=False),
        sa.ForeignKeyConstraint(
            ["mass_message_id"],
            ["onlyfans.x_mass_messages.id"],
        ),
        sa.ForeignKeyConstraint(
            ["media_id"],
            ["onlyfans.media.id"],
        ),
        sa.ForeignKeyConstraint(
            ["message_id"],
            ["onlyfans.x_messages.id"],
        ),
        sa.ForeignKeyConstraint(
            ["post_id"],
            ["onlyfans.x_posts.id"],
        ),
        sa.ForeignKeyConstraint(
            ["story_id"],
            ["onlyfans.x_stories.id"],
        ),
        sa.PrimaryKeyConstraint("id"),
        sa.UniqueConstraint("mass_message_id", "media_id"),
        sa.UniqueConstraint("message_id", "media_id"),
        sa.UniqueConstraint("post_id", "media_id"),
        sa.UniqueConstraint("story_id", "media_id"),
        schema="onlyfans",
    )
    # ### end Alembic commands ###


def downgrade() -> None:
    # ### commands auto generated by Alembic - please adjust! ###
    op.drop_table("filepaths", schema="onlyfans")
    op.drop_table("content_media_asso", schema="onlyfans")
    op.drop_table("filepaths", schema="fansly")
    op.drop_table("content_media_asso", schema="fansly")
    op.drop_table("x_messages", schema="onlyfans")
    op.drop_table("x_messages", schema="fansly")
    op.drop_table("x_mass_messages", schema="onlyfans")
    op.drop_table("media_detections", schema="onlyfans")
    op.drop_table("linked_users", schema="onlyfans")
    op.drop_table("linked_posts", schema="onlyfans")
    op.drop_table("comments", schema="onlyfans")
    op.drop_table("x_mass_messages", schema="fansly")
    op.drop_table("media_detections", schema="fansly")
    op.drop_table("linked_users", schema="fansly")
    op.drop_table("linked_posts", schema="fansly")
    op.drop_table("comments", schema="fansly")
    op.drop_table("x_stories", schema="onlyfans")
    op.drop_table("x_posts", schema="onlyfans")
    op.drop_index(
        op.f("ix_onlyfans_user_infos_sex"), table_name="user_infos", schema="onlyfans"
    )
    op.drop_table("user_infos", schema="onlyfans")
    op.drop_table("user_auths", schema="onlyfans")
    op.drop_table("user_aliases", schema="onlyfans")
    op.drop_table("subscriptions", schema="onlyfans")
    op.drop_table("socials", schema="onlyfans")
    op.drop_table("remote_urls", schema="onlyfans")
    op.drop_table("notifications", schema="onlyfans")
    op.drop_table("media", schema="onlyfans")
    op.drop_table("mass_message_stats", schema="onlyfans")
    op.drop_table("jobs", schema="onlyfans")
    op.drop_table("histo_user_infos", schema="onlyfans")
    op.drop_table("favorite_users", schema="onlyfans")
    op.drop_table("bought_content", schema="onlyfans")
    op.drop_table("x_stories", schema="fansly")
    op.drop_table("x_posts", schema="fansly")
    op.drop_index(
        op.f("ix_fansly_user_infos_sex"), table_name="user_infos", schema="fansly"
    )
    op.drop_table("user_infos", schema="fansly")
    op.drop_table("user_auths", schema="fansly")
    op.drop_table("user_aliases", schema="fansly")
    op.drop_table("subscriptions", schema="fansly")
    op.drop_table("socials", schema="fansly")
    op.drop_table("remote_urls", schema="fansly")
    op.drop_table("notifications", schema="fansly")
    op.drop_table("media", schema="fansly")
    op.drop_table("mass_message_stats", schema="fansly")
    op.drop_table("jobs", schema="fansly")
    op.drop_table("histo_user_infos", schema="fansly")
    op.drop_table("favorite_users", schema="fansly")
    op.drop_table("bought_content", schema="fansly")
    op.drop_table("users", schema="onlyfans")
    op.drop_table("sites", schema="management")
    op.drop_table("servers", schema="management")
    op.drop_table("hosts", schema="management")
    op.drop_table("users", schema="fansly")
    # ### end Alembic commands ###



from fastapi import APIRouter, Response
from pydantic import BaseModel

from ultima_scraper_db.databases.ultima_archive.api.client import UAClient
from ultima_scraper_db.databases.ultima_archive.schemas.management import HostModel

router = APIRouter(
    prefix="/hosts",
    tags=["hosts"],
    responses={404: {"description": "Not found"}},
)


@router.get("/")
async def get_hosts():
    database_api = UAClient.database_api

    async with database_api.create_management_api() as management_api:
        hosts = await management_api.get_hosts()
        return hosts


class PyHost(BaseModel):
    name: str
    identifier: str
    password: str
    source: bool
    active: bool


@router.post("/create")
async def create_host(response: Response, host: PyHost):
    database_api = UAClient.database_api

    async with database_api.create_management_api() as management_api:
        db_host = HostModel(
            name=host.name,
            identifier=host.identifier,
            password=host.password,
            source=host.source,
            active=host.active,
        )
        await management_api.create_or_update_host(db_host)
        return db_host



import asyncio
import gc
import weakref
from datetime import datetime, timedelta, timezone
from pathlib import Path
from types import TracebackType
from typing import TYPE_CHECKING, Any, Literal, Sequence, Type
from urllib.parse import ParseResult

import ultima_scraper_api
from alive_progress import alive_bar  # type: ignore
from inflection import underscore
from sqlalchemy import ScalarResult, Select, UnaryExpression, delete, or_, orm, select
from sqlalchemy.ext.asyncio import AsyncSession
from sqlalchemy.ext.asyncio.session import async_object_session
from sqlalchemy.orm import joinedload, noload, selectinload
from sqlalchemy.sql import and_, exists, func, select, union_all
from ultima_scraper_api.apis.fansly.classes.extras import (
    AuthDetails as FanslyAuthDetails,
)
from ultima_scraper_api.apis.onlyfans.classes.auth_model import OnlyFansAuthModel
from ultima_scraper_api.apis.onlyfans.classes.comment_model import (
    CommentModel as OFCommentModel,
)
from ultima_scraper_api.apis.onlyfans.classes.extras import (
    AuthDetails as OnlyFansAuthDetails,
)
from ultima_scraper_api.apis.onlyfans.classes.user_model import UserModel as OFUserModel
from ultima_scraper_api.helpers.main_helper import split_string
from ultima_scraper_collection.helpers.main_helper import is_notif_valuable
from ultima_scraper_collection.managers.aio_pika_wrapper import (
    AioPikaWrapper,
    create_notification,
)

from ultima_scraper_db.databases.ultima_archive.filters import AuthedInfoFilter
from ultima_scraper_db.databases.ultima_archive.schemas.management import SiteModel
from ultima_scraper_db.databases.ultima_archive.schemas.templates.site import (
    BoughtContentModel,
    CommentModel,
    ContentMediaAssoModel,
    ContentTemplate,
    FilePathModel,
    JobModel,
    JobStatus,
    MassMessageModel,
    MassMessageStatModel,
    MediaModel,
    MessageModel,
    NotificationModel,
    PostModel,
    RemoteURLModel,
    StoryModel,
    SubscriptionModel,
    UserAliasModel,
    UserAuthModel,
    UserInfoModel,
    UserModel,
)
from ultima_scraper_db.managers.database_manager import Schema

content_model_types = StoryModel | PostModel | MessageModel | MassMessageModel
from ultima_scraper_api import post_types

if TYPE_CHECKING:
    from ultima_scraper_collection import datascraper_types
    from ultima_scraper_collection.managers.metadata_manager.metadata_manager import (
        ContentMetadata,
        MediaMetadata,
    )

from sqlalchemy import inspect
from sqlalchemy.orm.strategy_options import _AbstractLoad  # type: ignore


def create_options(
    alias: bool = False,
    user_info: bool = False,
    content: bool = False,
    media: bool = False,
    notifications: bool = False,
    remote_urls: bool = False,
):
    joined_options: list[_AbstractLoad] = []
    if alias:
        stmt = selectinload(UserModel.aliases)
        joined_options.append(stmt)
    if content:
        inspector = inspect(UserModel)
        for relationship_name, relationship in inspector.mapper.relationships.items():
            if issubclass(relationship.mapper.class_, ContentTemplate):
                relationship_attribute = getattr(UserModel, relationship_name)
                inspector2 = inspect(relationship_attribute)
                content_relationship = inspector2.mapper.relationships
                stmt = selectinload(relationship)  # type: ignore
                if hasattr(content_relationship, "media"):
                    stmt = stmt.selectinload(
                        content_relationship["media"]
                    ).selectinload(MediaModel.filepaths)
                joined_options.append(stmt)
    if media:
        stmt = selectinload(UserModel.medias).options(
            selectinload(MediaModel.filepaths),
            selectinload(MediaModel.content_media_assos),
        )
        joined_options.append(stmt)
    if user_info:
        stmt = selectinload(UserModel.user_info)
        joined_options.append(stmt)
    if notifications:
        stmt = selectinload(UserModel.notifications)
        joined_options.append(stmt)
    if remote_urls:
        stmt = selectinload(UserModel.remote_urls)
        joined_options.append(stmt)
    return joined_options


def fix_missing_paid_content(
    db_content: PostModel | MessageModel,
    content: "ContentMetadata",
    datascraper: "datascraper_types",
    api_performer: ultima_scraper_api.user_types,
):
    if bool(db_content.paid) == False and content.price and content.price > 0:
        fmu = datascraper.filesystem_manager.get_file_manager(api_performer.id)
        valid_local_media_count = 0
        for file in fmu.files:
            remote_media_date = content.__soft__.created_at
            remote_post_date = content.__soft__.created_at
            stat = file.stat()
            mt = stat.st_mtime
            local_media_date = datetime.fromtimestamp(mt).replace(microsecond=0)
            local_post_date = datetime.fromtimestamp(mt).replace(microsecond=0)
            if (
                local_media_date == remote_media_date
                or local_post_date == remote_post_date
            ):
                valid_local_media_count += 1
        if valid_local_media_count == content.__soft__.media_count:
            db_content.paid = True


class FilePathManager:
    def __init__(self, content_manager: "ContentManager") -> None:
        self.content_manager = content_manager
        self.filepaths: dict[int, FilePathModel] = {
            filepath.id: filepath
            for media in content_manager.media_manager.medias.values()
            for filepath in media.filepaths
        }
        self._filepaths_str: dict[str, FilePathModel] = {
            Path(item.filepath).name: item for item in self.filepaths.values()
        }

    def resolve_filepath(self, identifier: int | str):
        filepath = self.filepaths.get(identifier)
        if not filepath:
            filepath = self._filepaths_str.get(identifier)
        return filepath


class MediaManager:
    def __init__(self, content_manager: "ContentManager") -> None:
        self.content_manager = content_manager
        self.medias: dict[int, MediaModel] = {}

    async def init(self, media_ids: list[int] = []):
        content_manager = self.content_manager
        user = content_manager.__user__
        for media in user.medias:
            self.medias[media.id] = media
        # remove duplicate ids from media_ids by self.media_ids
        final_media_ids = [x for x in media_ids if x not in self.medias]
        if media_ids:
            stmt = (
                select(MediaModel)
                .where(MediaModel.id.in_(final_media_ids))
                .options(
                    joinedload(MediaModel.filepaths),
                    joinedload(MediaModel.content_media_assos),
                )
            )
            assert content_manager.session
            found_medias = await content_manager.session.scalars(stmt)
            for media in found_medias.unique():
                self.medias[media.id] = media
        self.filepath_manager = FilePathManager(content_manager)

    def find_media(self, media_id: int):
        try:
            return self.medias[media_id]
        except KeyError:
            return

    def add_media(self, media_model: MediaModel):
        self.medias[media_model.id] = media_model


class ContentManager:
    def __init__(
        self,
        user: "UserModel",
    ):
        self.__user__ = user
        self.session = async_object_session(user)
        self.lock = asyncio.Lock()
        self.media_manager = MediaManager(self)

    async def init(self, load_media: bool = False, media_ids: list[int] = []):
        awaitables = [
            self.__user__.awaitable_attrs._stories,
            self.__user__.awaitable_attrs._posts,
            self.__user__.awaitable_attrs._messages,
            self.__user__.awaitable_attrs._mass_messages,
        ]
        await asyncio.gather(*awaitables)

        if not self.session:
            self.session = async_object_session(self.__user__)

        media_manager = self.media_manager
        self.stories: list["StoryModel"] = self.__user__._stories  # type: ignore
        stories_media = [content.awaitable_attrs.media for content in self.stories]

        self.posts: list["PostModel"] = self.__user__._posts  # type: ignore
        posts_media = [content.awaitable_attrs.media for content in self.posts]

        self.messages: list["MessageModel"] = self.__user__._messages  # type: ignore
        messages_media = [content.awaitable_attrs.media for content in self.messages]

        self.mass_messages: list["MassMessageModel"] = self.__user__._mass_messages  # type: ignore
        mass_messages_media = [
            content.awaitable_attrs.media for content in self.mass_messages
        ]

        if load_media:
            stories_media_items = await asyncio.gather(*stories_media)
            for media_list in stories_media_items:
                for item in media_list:
                    media_manager.add_media(item)
            posts_media_items = await asyncio.gather(*posts_media)
            for media_list in posts_media_items:
                for item in media_list:
                    media_manager.add_media(item)
            messages_media_items = await asyncio.gather(*messages_media)
            for media_list in messages_media_items:
                for item in media_list:
                    media_manager.add_media(item)
            mass_messages_media_items = await asyncio.gather(*mass_messages_media)
            for media_list in mass_messages_media_items:
                for item in media_list:
                    media_manager.add_media(item)
            await media_manager.init(media_ids=media_ids)
        return self

    def get_media_manager(self):
        return self.media_manager

    def get_filepath_manager(self):
        return self.get_media_manager().filepath_manager

    async def get_contents(self, content_type: str | None = None):
        if content_type:
            # if not hasattr(self, content_type.lower()):
            #     empty_list: content_model_types = []
            #     return empty_list
            result: list[content_model_types] = getattr(
                self, underscore(content_type).lower()
            )
            return result
        return self.stories + self.posts + self.messages + self.mass_messages

    async def add_content(self, content: "ContentMetadata"):
        match content.api_type:
            case "Stories":
                content_model = StoryModel(
                    id=content.content_id,
                    user_id=content.user_id,
                    created_at=content.__soft__.created_at,
                )
                self.stories.append(content_model)
            case "Posts" | "Archived/Posts":
                content_model = PostModel(
                    id=content.content_id,
                    user_id=content.user_id,
                    text=content.text,
                    price=content.price,
                    paid=int(bool(content.paid)),
                    created_at=content.__soft__.created_at,
                )
                self.posts.append(content_model)
            case "Messages":
                queue_id = (
                    content.queue_id if content.__soft__.is_mass_message() else None
                )
                content_model = MessageModel(
                    id=content.content_id,
                    user_id=content.user_id,
                    receiver_id=content.receiver_id,
                    text=content.text,
                    price=content.price,
                    paid=int(bool(content.paid)),
                    verified=True,
                    queue_id=queue_id,
                    created_at=content.__soft__.created_at,
                )
                self.messages.append(content_model)
                assert content.receiver_id
                content_model.receiver_id = content.receiver_id
                pass
            case "MassMessages":
                stat = content.get_mass_message_stat()
                content_model = MassMessageModel(
                    id=content.content_id,
                    user_id=content.user_id,
                    statistic_id=stat.id if stat else None,
                    text=content.text,
                    price=content.price,
                    expires_at=content.__soft__.expires_at,
                    created_at=content.__soft__.created_at,
                )
                self.mass_messages.append(content_model)
            case _:
                raise Exception("Content not assigned")
        return content_model

    async def find_content(
        self,
        content_id: int,
        content_type: str | None = None,
    ):
        temp_contents = await self.get_contents(content_type=content_type)
        for content in temp_contents:
            if content.id == content_id:
                return content

    async def find_paid_contents(
        self, only_downloaded: bool = False
    ) -> Sequence[MediaModel]:
        session = self.session
        assert session

        # Adjust common conditions based on only_downloaded parameter
        if only_downloaded:
            common_conditions = and_(
                MediaModel.filepaths.any(
                    FilePathModel.downloaded == True
                ),  # Assuming MediaFilePath is the model for filepaths
                MediaModel.preview == False,
            )
        else:
            common_conditions = and_(
                MediaModel.filepaths.any(), MediaModel.preview == False
            )

        # PostModel conditions
        post_conditions = and_(
            PostModel.user_id == self.__user__.id,
            PostModel.price > 0,
            PostModel.paid == True,
        )

        # MessageModel conditions
        message_conditions = and_(
            MessageModel.user_id == self.__user__.id,
            MessageModel.price > 0,
            MessageModel.paid == True,
        )

        stmt1 = (
            select(MediaModel)
            .join(PostModel.media)
            .options(selectinload(MediaModel.filepaths))
            .where(post_conditions, common_conditions)
        )

        stmt2 = (
            select(MediaModel)
            .join(MessageModel.media)
            .options(selectinload(MediaModel.filepaths))
            .where(message_conditions, common_conditions)
        )

        union_stmt = union_all(stmt1, stmt2)
        orm_stmt = select(MediaModel).from_statement(union_stmt)
        result = await session.scalars(orm_stmt)
        return result.unique().all()

    async def size_sum(self):
        session = self.session
        assert session

        stmt = select(func.sum(MediaModel.size)).where(
            MediaModel.user_id == self.__user__.id
        )

        result = await session.scalar(stmt)
        final_sum = int(result or 0)
        return final_sum

    async def media_sum(self, category: str):
        session = async_object_session(self.__user__)
        assert session

        stmt1 = (
            select(func.count())
            .where(MediaModel.category == category)
            .where(MediaModel.filepaths.any())
            .join(StoryModel.media)
            .where(StoryModel.user_id == self.__user__.id)
        )
        stmt2 = (
            select(func.count())
            .where(MediaModel.category == category)
            .where(MediaModel.filepaths.any())
            .join(PostModel.media)
            .where(PostModel.user_id == self.__user__.id)
        )
        stmt3 = (
            select(func.count())
            .where(MediaModel.category == category)
            .where(MediaModel.filepaths.any())
            .join(MessageModel.media)
            .where(MessageModel.user_id == self.__user__.id)
        )
        result1 = await session.scalar(stmt1)
        result2 = await session.scalar(stmt2)
        result3 = await session.scalar(stmt3)
        final_sum = sum((result1 or 0, result2 or 0, result3 or 0))
        return final_sum


class StatementBuilder:
    def __init__(self, model: Type[UserModel | UserInfoModel]) -> None:
        self.model = model
        self.statement: Select[Any] = select(model)

    def filter_by_user_identifiers(
        self, identifiers: list[int | str] | int | str | None
    ):
        if not identifiers:
            return self
        if isinstance(identifiers, str):
            identifiers = identifiers.replace("@", "")
            identifiers = identifiers.replace(" ", ",")
            final_identifiers = split_string(identifiers)
        else:
            if isinstance(identifiers, int):
                identifiers = [identifiers]
            final_identifiers = [str(x) for x in identifiers]
        template_query = [
            UserModel.username.in_([x for x in final_identifiers if not x.isdigit()]),
            UserModel.id.in_([int(x) for x in final_identifiers if x.isdigit()]),
        ]
        if self.model == UserModel:
            self.statement = self.statement.filter(or_(*template_query))
        else:
            self.statement = self.statement.join(UserModel).filter(or_(*template_query))

        return self

    def filter_by_description(self, description: str | None):
        if not description:
            return self
        description = description.lower()
        if self.model == UserModel:
            self.statement = self.statement.join(UserInfoModel).filter(
                UserInfoModel.description.ilike(f"%{description}%")
            )
        else:
            self.statement = self.statement.join(UserModel).filter(
                UserInfoModel.description.ilike(f"%{description}%")
            )
        return self


class SiteAPI:
    def __init__(
        self,
        schema: Schema,
        datascraper: "datascraper_types | None" = None,
    ) -> None:
        self.database = schema.database
        self.schema = schema
        self.datascraper = datascraper
        self.content_managers: dict[int, "weakref.ref[ContentManager]"] = {}
        self.aio_pika_wrapper: AioPikaWrapper | None = None

    async def __aenter__(self):
        self._session: AsyncSession = self.schema.sessionmaker()
        return self

    async def __aexit__(
        self,
        exc_type: None | Type[BaseException],
        exc_value: None | BaseException,
        traceback: None | TracebackType,
    ):
        if exc_type:
            await self._session.rollback()
        else:
            await self._session.commit()
        await self._session.close()
        self.content_managers.clear()
        gc.collect()

    def resolve_session(self):
        assert self._session, "Session has not been set"
        return self._session

    def get_session(self):
        return self.resolve_session()

    def set_session(self, session: AsyncSession):
        self._session = session

    def get_session_maker(self):
        return self.schema.sessionmaker()

    def resolve_content_manager(self, user: UserModel):
        content_manager = self.content_managers.get(user.id)
        if content_manager:
            content_manager = content_manager()
        if not content_manager:
            content_manager = user.content_manager
            if not content_manager:
                content_manager = ContentManager(user)
                user.content_manager = content_manager
            self.content_managers[user.id] = weakref.ref(content_manager)

        return content_manager

    def get_user_query(
        self,
        identifiers: list[int | str] | int | str | None = None,
        description: str | None = None,
        performer: bool | None = None,
        has_paid_content: bool | None = None,
        load_aliases: bool = False,
        load_user_info: bool = False,
        load_content: bool = False,
        load_media: bool = False,
        load_notifications: bool = False,
        load_remote_urls: bool = False,
        limit: int | None = None,
        extra_options: Any = (),
    ):
        options = create_options(
            content=load_content,
            media=load_media,
            user_info=load_user_info,
            alias=load_aliases,
            notifications=load_notifications,
            remote_urls=load_remote_urls,
        )
        options += extra_options
        stmt_builder = (
            StatementBuilder(UserModel)
            .filter_by_user_identifiers(identifiers)
            .filter_by_description(description)
        )
        stmt = stmt_builder.statement.options(*options).distinct().limit(limit)
        if performer is not None:
            stmt = stmt.where(UserModel.performer == performer)
        if has_paid_content is not None:
            stmt = (
                stmt.where(UserModel.supplied_contents.any())
                .join(UserModel.supplied_contents)
                .where(UserAuthModel.active == True)
            )
        return stmt

    async def get_users(
        self,
        identifiers: list[int | str] | str | None = None,
        description: str | None = None,
        performer: bool | None = None,
        has_paid_content: bool | None = None,
        load_aliases: bool = False,
        load_user_info: bool = True,
        load_content: bool = False,
        load_media: bool = False,
        load_notifications: bool = False,
        load_remote_urls: bool = False,
        limit: int | None = None,
        extra_options: Any = (),
        authed_info_filter: AuthedInfoFilter | None = None,
        order_by: UnaryExpression[Any] | None = None,
    ):
        if identifiers:
            if isinstance(identifiers, str):
                identifiers = [identifiers]
            for i, identifier in enumerate(identifiers):
                if (
                    identifier
                    and isinstance(identifier, str)
                    and identifier[0].lower() == "u"
                ):
                    temp_identifier = identifier[1:]
                    if temp_identifier.isdigit():
                        identifiers[i] = int(temp_identifier)
        stmt = self.get_user_query(
            identifiers,
            description,
            performer,
            has_paid_content,
            load_aliases,
            load_user_info,
            load_content,
            load_media,
            load_notifications,
            load_remote_urls,
            limit,
            extra_options,
        )
        if authed_info_filter:
            if authed_info_filter.exclude_between_dates:
                stmt = stmt.where(
                    ~UserModel.user_auths_info.any(
                        UserModel.last_checked_at.between(
                            *authed_info_filter.exclude_between_dates
                        )
                    )
                )
            if authed_info_filter.active is not None:
                stmt = stmt.where(
                    UserModel.user_auths_info.any(active=authed_info_filter.active)
                )
        if order_by is not None:
            stmt = stmt.order_by(order_by)
        session = self.get_session()
        result: ScalarResult[UserModel] = await session.scalars(stmt)
        db_users = result.all()
        db_users = list(set(db_users))
        for db_user in db_users:
            if db_user and load_content:
                await self.resolve_content_manager(db_user).init(load_media=load_media)
        if isinstance(identifiers, list):
            invalid_identifiers: list[str] = []
            for identifier in identifiers:
                if isinstance(identifier, str):
                    found_db_user = None
                    for db_user in db_users:
                        found_db_user = await db_user.find_username(identifier)
                        if found_db_user:
                            break
                    if not found_db_user:
                        invalid_identifiers.append(identifier)
            for invalid_identifier in invalid_identifiers:
                stmt = select(UserAliasModel).where(
                    UserAliasModel.username == invalid_identifier
                )
                db_alias = await session.scalar(stmt)
                if db_alias:
                    db_user = await self.get_user(
                        db_alias.user_id,
                        load_user_info=load_user_info,
                        load_aliases=load_aliases,
                        load_content=load_content,
                        load_remote_urls=load_remote_urls,
                        extra_options=extra_options,
                    )
                    if db_user:
                        db_users.append(db_user)
        return db_users

    async def get_user(
        self,
        identifier: int | str | None,
        description: str | None = None,
        performer: bool | None = None,
        has_paid_content: bool | None = None,
        load_aliases: bool = False,
        load_user_info: bool = False,
        load_content: bool = False,
        load_media: bool = False,
        load_notifications: bool = False,
        load_remote_urls: bool = False,
        limit: int | None = None,
        extra_options: Any = (),
    ) -> UserModel | None:
        if identifier and isinstance(identifier, str) and identifier[0].lower() == "u":
            temp_identifier = identifier[1:]
            if temp_identifier.isdigit():
                identifier = int(temp_identifier)
        stmt = self.get_user_query(
            identifier,
            description,
            performer,
            has_paid_content,
            load_aliases,
            load_user_info,
            load_content,
            load_media,
            load_notifications,
            load_remote_urls,
            limit,
            extra_options,
        )
        session = self.get_session()
        result: ScalarResult[UserModel] = await session.scalars(stmt)
        db_user = result.first()
        if db_user and load_content:
            await self.resolve_content_manager(db_user).init(load_media=load_media)

        if not db_user and isinstance(identifier, str):
            stmt = select(UserAliasModel).where(UserAliasModel.username == identifier)
            db_alias = await session.scalar(stmt)
            if db_alias:
                db_user = await self.get_user(
                    db_alias.user_id,
                    load_user_info=load_user_info,
                    load_aliases=load_aliases,
                    load_content=load_content,
                    load_remote_urls=load_remote_urls,
                    extra_options=extra_options,
                )
        return db_user

    async def get_subscription(self, user_id: int, subscriber_id: int):
        session = self.get_session()
        stmt = select(SubscriptionModel).filter_by(
            user_id=user_id, subscriber_id=subscriber_id
        )
        found_subscription = await session.scalar(stmt)
        return found_subscription

    async def get_post(self, post_id: int):
        stmt = select(PostModel).where(PostModel.id == post_id)

        found_post = await self.get_session().scalars(stmt)
        return found_post.first()

    async def get_posts(
        self,
        user_id: int | None = None,
        post_id: int | None = None,
        text: str | None = None,
        ppv: bool | None = None,
        paid: bool | None = None,
        load_media: bool = True,
    ):
        stmt = select(PostModel)
        if user_id is not None:
            stmt = stmt.where(PostModel.user_id == user_id)
        if ppv is not None:
            stmt = stmt.where(PostModel.price > 0)
        if paid is not None:
            stmt = stmt.where(PostModel.paid == paid)
        if text is not None:
            text_list = text.split(",")
            for text_item in text_list:
                text_item = text_item.strip()
                stmt = stmt.where(PostModel.text.icontains(text_item))

        options: list[_AbstractLoad] = [joinedload(PostModel.user)]
        if not load_media:
            options.append(noload(PostModel.media))
        stmt = stmt.options(*options)
        found_posts = await self.get_session().scalars(stmt)
        return found_posts.unique().all()

    async def get_messages(
        self,
        user_id: int | None = None,
        message_id: int | None = None,
        text: str | None = None,
        ppv: bool | None = None,
        paid: bool | None = None,
        load_media: bool = True,
    ):
        stmt = select(MessageModel)
        if user_id is not None:
            stmt = stmt.where(MessageModel.user_id == user_id)
        if ppv is not None:
            stmt = stmt.where(MessageModel.price > 0)
        if paid is not None:
            stmt = stmt.where(MessageModel.paid == paid)
        if text is not None:
            text_list = text.split(",")
            for text_item in text_list:
                text_item = text_item.strip()
                stmt = stmt.where(MessageModel.text.icontains(text_item))
        options: list[_AbstractLoad] = [joinedload(MessageModel.user)]
        if not load_media:
            options.append(noload(MessageModel.media))
        stmt = stmt.options(*options)
        found_contents = await self.get_session().scalars(stmt)
        return found_contents.all()

    async def get_media(self, media_id: int | None = None, url: str | None = None):
        found_media = None
        stmt = select(MediaModel)
        if media_id:
            stmt = stmt.where(MediaModel.id == media_id)
        if url:
            stmt = stmt.where(MediaModel.url.contains(url))
        found_media = await self.get_session().scalar(stmt)
        if found_media:
            await found_media.awaitable_attrs.filepaths
        return found_media

    async def get_medias(self, user_id: int, media_ids: list[int] | None = None):
        stmt = (
            select(MediaModel)
            .where(MediaModel.user_id == user_id)
            .options(
                joinedload(MediaModel.filepaths),
                joinedload(MediaModel.content_media_assos),
            )
        )
        if media_ids is not None:
            stmt = stmt.where(MediaModel.id.in_(media_ids))
        found_media = await self.get_session().scalars(stmt)
        return found_media.unique().all()

    async def get_filepaths(
        self,
        identifier: int | str | None = None,
        media_id: int | None = None,
        post_id: int | None = None,
    ):
        stmt = select(FilePathModel)

        if identifier is not None:
            if isinstance(identifier, int):
                stmt = stmt.where(FilePathModel.id == identifier)
            else:
                stmt = stmt.where(FilePathModel.filepath.contains(identifier))

        if media_id is not None:
            stmt = stmt.where(FilePathModel.media_id == media_id)

        if post_id is not None:
            stmt = stmt.join(MediaModel).join(PostModel).where(PostModel.id == post_id)

        found_filepaths = await self.get_session().scalars(stmt)
        return found_filepaths.all()

    async def get_mass_message(self, mass_message_id: int):
        stmt = select(MassMessageModel).where(MassMessageModel.id == mass_message_id)
        found_mass_message = await self.get_session().scalars(stmt)
        return found_mass_message.first()

    async def get_site(self):
        stmt = select(SiteModel).filter_by(db_name=self.schema.name)
        found_site = await self.get_session().scalar(stmt)
        assert found_site
        return found_site

    async def get_notifications(
        self,
        platform: Literal["discord", "telegram"] | None = None,
        sent: bool | None = None,
        page: int = 1,
        limit: int = 100,
    ):
        stmt = select(NotificationModel).options(
            joinedload(NotificationModel.user).joinedload(UserModel.user_info),
            joinedload(NotificationModel.user).joinedload(UserModel.aliases),
        )
        if platform and sent is not None:
            if platform == "telegram":
                stmt = stmt.where(NotificationModel.sent_telegram == sent)
            else:
                stmt = stmt.where(NotificationModel.sent_discord == sent)
        # Calculate offset
        offset = (page - 1) * limit

        # Apply limit and offset for pagination
        stmt = stmt.limit(limit).offset(offset)

        found_notifications = await self.get_session().scalars(stmt)
        return found_notifications.unique().all()

    async def get_jobs(
        self,
        server_id: int | None = None,
        performer_id: int | None = None,
        user_id: int | None = None,
        category: str | None = None,
        priority: int | None = None,
        status: JobStatus | None = None,
        active: bool | None = None,
        page: int = 1,
        limit: int | None = 100,
    ):
        session = self.get_session()
        db_site = await self.get_site()

        stmt = select(JobModel).filter_by(site_id=db_site.id)
        if server_id:
            stmt = stmt.filter_by(server_id=server_id)
        if performer_id:
            stmt = stmt.filter_by(user_id=performer_id)
        if user_id:
            stmt = stmt.filter_by(user_id=user_id)
        if category:
            stmt = stmt.filter_by(category=category)
        if status:
            stmt = stmt.filter_by(status=status)
        if priority is not None:
            stmt = stmt.filter_by(priority=priority)
        if active is not None:
            stmt = stmt.filter_by(active=active)

        stmt = (
            stmt.join(JobModel.user)
            .order_by(JobModel.priority.desc())
            .order_by(JobModel.id.asc())
            .order_by(UserModel.downloaded_at.desc())
            .options(orm.contains_eager(JobModel.user))
        )
        if limit:
            offset = max(0, (page - 1) * limit)
            stmt = stmt.offset(offset).limit(limit)

        found_jobs = await session.scalars(stmt)
        return found_jobs.all()

    async def create_or_update_job(
        self,
        db_user: UserModel,
        category: str,
        options: dict[str, Any] | None = None,
        server_id: int = 1,
        host_id: int | None = None,
        priority: int = 0,
        skippable: bool = False,
    ):
        session = self.get_session()
        db_jobs = await self.get_jobs(user_id=db_user.id, category=category)
        db_site = await self.get_site()
        if db_jobs:
            db_job = db_jobs[0]
            db_job.site_id = db_site.id
            db_job.user_id = db_user.id
            db_job.user_username = db_user.username
            db_job.category = category
            db_job.options = options
            db_job.server_id = server_id
            db_job.priority = priority
            db_job.host_id = host_id
            db_job.skippable = skippable
            db_job.active = True
            db_job.status = (
                JobStatus.QUEUED
                if db_job.status != JobStatus.RUNNING
                else db_job.status
            )
        else:
            db_job = JobModel(
                site_id=db_site.id,
                user_id=db_user.id,
                user_username=db_user.username,
                category=category,
                options=options,
                server_id=server_id,
                host_id=host_id,
                skippable=skippable,
                priority=priority,
            )
            session.add(db_job)
        await session.commit()
        return db_job

    async def fetch_next_job(self, categories: list[str] | None = None):
        session = self.resolve_session()

        now = datetime.now(timezone.utc)
        timeout_threshold = now - timedelta(minutes=5)

        stmt = select(JobModel).where(
            or_(
                JobModel.status.in_([JobStatus.QUEUED, JobStatus.RETRYING]),
                and_(
                    JobModel.status == JobStatus.RUNNING,
                    or_(
                        JobModel.heartbeat_at == None,  # NULL heartbeat means stale
                        JobModel.heartbeat_at < timeout_threshold,
                    ),
                ),
            )
        )

        if categories:
            stmt = stmt.where(JobModel.category.in_(categories))

        stmt = (
            stmt.order_by(
                JobModel.priority.desc(), JobModel.id.asc(), JobModel.heartbeat_at.asc()
            )
            .with_for_update(skip_locked=True)
            .limit(1)
        )

        result = await session.execute(stmt)
        job = result.scalar_one_or_none()

        if job:
            job.status = JobStatus.RUNNING
            job.heartbeat_at = now
            await session.commit()

        return job

    async def get_remote_urls(self, user_id: int):
        stmt = select(RemoteURLModel).where(RemoteURLModel.user_id == user_id)
        found_remote_urls = await self.get_session().scalars(stmt)
        return found_remote_urls.all()

    async def get_bulk_remote_urls(self, user_ids: list[int]):
        stmt = select(RemoteURLModel).where(RemoteURLModel.user_id.in_(user_ids))
        found_remote_urls = await self.get_session().scalars(stmt)
        return found_remote_urls.all()

    async def create_or_update_remote_url(
        self, user_id: int, host_id: int, url: str, part: int, uploaded_at: datetime
    ):
        session = self.get_session()
        stmt = (
            select(RemoteURLModel)
            .where(RemoteURLModel.user_id == user_id)
            .where(RemoteURLModel.part == part)
        )
        found_remote_url = await session.scalar(stmt)
        if not found_remote_url:
            found_remote_url = RemoteURLModel(
                user_id=user_id,
                host_id=host_id,
                url=url,
                part=part,
                uploaded_at=uploaded_at,
                exists=True,
            )
            session.add(found_remote_url)
        else:
            found_remote_url.host_id = host_id
            found_remote_url.url = url
            found_remote_url.uploaded_at = uploaded_at
            found_remote_url.exists = True
        await session.commit()
        return found_remote_url

    async def update_user(
        self, api_user: ultima_scraper_api.user_types, found_db_user: UserModel | None
    ):
        assert self.datascraper
        assert found_db_user

        content_manager = self.datascraper.resolve_content_manager(api_user)
        for media in content_manager.media_manager.medias.values():
            _db_media = await self.create_or_update_media(found_db_user, media)
            _db_filepath = await self.create_or_update_filepaths(found_db_user, media)
        _db_user = await self.create_or_update_user(
            api_user, existing_user=found_db_user, performer_optimize=True
        )
        current_job = api_user.get_current_job()
        if current_job:
            assert current_job
            current_job.done = True

    async def create_or_update_user(
        self,
        api_user: ultima_scraper_api.user_types,
        existing_user: UserModel | None,
        performer_optimize: bool = False,
        update_socials: bool = True,
        skip_subscriptions: bool = False,
        skip_paid_content: bool = False,
    ):
        session = self.get_session()
        db_user = existing_user or UserModel()
        db_user.id = api_user.id
        db_user.username = api_user.username
        db_user.balance = api_user.credit_balance or 0
        db_user.performer = api_user.is_performer()
        db_user.join_date = (
            datetime.fromisoformat(api_user.join_date) if api_user.join_date else None
        )
        if not existing_user:
            session.add(db_user)
            await self.resolve_content_manager(db_user).init()
        try:
            await session.commit()
        except Exception as e:
            pass
        await db_user.awaitable_attrs.medias
        await self.create_or_update_user_info(api_user, db_user)
        await db_user.update_username(api_user.username)
        status = False
        if not existing_user:
            await db_user.awaitable_attrs.user_auths_info
            if await is_notif_valuable(api_user):
                status = True
        if existing_user:
            await db_user.awaitable_attrs.subscribers
            if not db_user.subscribers:
                if await is_notif_valuable(api_user):
                    status = True
        if (
            api_user.is_authed_user()
            and api_user.is_performer()
            and not db_user.user_auths_info
        ):
            status = True

        if status:
            if db_user.last_checked_at == None:
                if self.aio_pika_wrapper:
                    message = create_notification(
                        "new_performer", api_user.get_api().site_name, db_user
                    )
                    await self.aio_pika_wrapper.publish_notification(message)
                else:
                    await db_user.awaitable_attrs.notifications
                    notification_exists = [
                        x
                        for x in db_user.notifications
                        if x.category == "new_performer"
                    ]
                    if not notification_exists:
                        notification = NotificationModel(
                            user_id=api_user.id, category="new_performer"
                        )
                        db_user.notifications.append(notification)
        if api_user.is_authed_user():
            api_authed = api_user.get_authed()
            db_auth_info = await self.create_or_update_auth_info(api_authed, db_user)
            if api_authed.is_authed():
                await db_auth_info.activate()
                if api_authed.user.is_performer():
                    if isinstance(api_authed, OnlyFansAuthModel):
                        mass_message_stats = await api_authed.get_mass_message_stats()
                        mass_message_stat_collection: dict[
                            int, MassMessageStatModel
                        ] = {}
                        for mass_message_stat in mass_message_stats:
                            found_mass_message = await db_user.find_mass_message(
                                mass_message_stat.id
                            )
                            if not found_mass_message:
                                media_types = mass_message_stat.media_types
                                media_count = (
                                    sum(media_types.values()) if media_types else 0
                                )
                                purchased_count = mass_message_stat.purchased_count
                                price = mass_message_stat.price
                                db_mass_message_stat = mass_message_stat_collection.get(
                                    mass_message_stat.id
                                )
                                if not db_mass_message_stat:
                                    db_mass_message_stat = MassMessageStatModel(
                                        id=mass_message_stat.id,
                                        user_id=db_user.id,
                                        media_count=media_count,
                                        buyer_count=purchased_count,
                                        sent_count=mass_message_stat.sent_count,
                                        view_count=mass_message_stat.viewed_count,
                                    )
                                db_mass_message = MassMessageModel(
                                    id=db_mass_message_stat.id,
                                    user_id=db_user.id,
                                    text=mass_message_stat.text,
                                    price=price,
                                    expires_at=mass_message_stat.expires_at,
                                    created_at=mass_message_stat.created_at,
                                    mass_message_stat=db_mass_message_stat,
                                )
                                db_user._mass_messages.append(db_mass_message)
                            pass
                if not skip_paid_content:
                    await self.create_or_update_paid_content(api_authed, db_user, [])
            if not skip_subscriptions:
                await self.process_subscriptions(
                    api_authed, db_user, performer_optimize
                )
        if isinstance(api_user, OFUserModel):
            status = True
            if performer_optimize and api_user.is_performer() or not update_socials:
                status = False
            if status:
                socials = await api_user.get_socials()
                await db_user.add_socials(socials)

                spotify = await api_user.get_spotify()
                if spotify:
                    spotify["socialMedia"] = "spotify"
                    spotify["username"] = spotify["displayName"]
                    await db_user.add_socials([spotify])
        await self.process_content(db_user, api_user)
        db_user.last_checked_at = datetime.now()
        await session.commit()
        return db_user

    async def create_or_update_auth_info(
        self, api_authed: ultima_scraper_api.auth_types, db_user: UserModel
    ):
        await db_user.awaitable_attrs.user_auths_info
        db_auth_info: UserAuthModel | None = None

        auth_info = api_authed.get_auth_details()
        for db_auth_info in db_user.user_auths_info:
            AD = db_auth_info.convert_to_auth_details(api_authed.get_api().site_name)
            if isinstance(auth_info, OnlyFansAuthDetails):
                assert isinstance(AD, OnlyFansAuthDetails)
                if auth_info.cookie.sess == AD.cookie.sess:
                    break
            else:
                FYD = db_auth_info.convert_to_auth_details(
                    api_authed.get_api().site_name
                )
                assert isinstance(FYD, FanslyAuthDetails)
                if auth_info.authorization == FYD.authorization:
                    break

        if not db_auth_info:
            exported_auth_details = api_authed.get_auth_details().export(UserAuthModel)
            db_auth_info = UserAuthModel(**exported_auth_details)
            db_auth_info.active = db_auth_info.active
            db_user.user_auths_info.append(db_auth_info)
            pass
        else:
            exported_auth_details = api_authed.get_auth_details().export()
            db_auth_info.update(exported_auth_details)
            pass
        return db_auth_info

    async def create_or_update_subscription(
        self,
        subscription: ultima_scraper_api.subscription_types,
        db_authed: "UserModel",
        performer_optimize: bool = False,
    ):
        db_sub_user = await self.get_user(
            subscription.user.id, load_content=True, load_media=True
        )
        db_sub_user = await self.create_or_update_user(
            subscription.user,
            db_sub_user,
            performer_optimize=performer_optimize,
        )
        subscription_user = subscription.user
        db_subscription = await self.get_subscription(
            subscription_user.id, db_authed.id
        )

        authed = subscription_user.get_authed()

        if not db_subscription:
            db_subscription = SubscriptionModel()
            await db_sub_user.awaitable_attrs.subscribers
            db_sub_user.subscribers.append(db_subscription)

        # Common lines of code for both if and else cases
        db_subscription.user_id = subscription_user.id
        db_subscription.subscriber_id = authed.id
        if isinstance(subscription_user, OFUserModel):
            db_subscription.paid_content = bool(
                await subscription_user.get_paid_contents()
            )
        db_subscription.created_at = datetime.fromisoformat(
            subscription.subscribed_by_data["subscribeAt"]
        )
        db_subscription.renewed_at = (
            datetime.fromisoformat(subscription.subscribed_by_data["renewedAt"])
            if subscription.subscribed_by_data["renewedAt"]
            else None
        )
        db_subscription.expires_at = subscription.resolve_expires_at()
        db_subscription.active = subscription.is_active()
        await self.get_session().commit()
        return db_subscription

    async def create_or_update_user_info(
        self,
        subscription_user: ultima_scraper_api.user_types,
        db_user: "UserModel",
    ):
        from ultima_scraper_api.apis.onlyfans.classes.user_model import UserModel

        await db_user.awaitable_attrs.user_info
        if not db_user.user_info:
            db_user.user_info = UserInfoModel()
        user_info = db_user.user_info
        user_info.name = subscription_user.name
        user_info.description = subscription_user.about
        user_info.price = await subscription_user.subscription_price() or 0
        user_info.post_count = subscription_user.posts_count
        user_info.media_count = subscription_user.medias_count
        user_info.image_count = subscription_user.photos_count
        user_info.video_count = subscription_user.videos_count
        user_info.audio_count = subscription_user.audios_count
        if isinstance(subscription_user, UserModel):
            user_info.stream_count = subscription_user.finished_streams_count
        user_info.archived_post_count = subscription_user.archived_posts_count
        if isinstance(subscription_user, UserModel):
            user_info.private_archived_post_count = (
                subscription_user.private_archived_posts_count
            )
        user_info.favourited_count = subscription_user.favorited_count
        if isinstance(subscription_user, UserModel):
            user_info.favourites_count = subscription_user.favorites_count
        user_info.subscribers_count = subscription_user.subscribers_count or 0
        user_info.location = subscription_user.location
        user_info.website = subscription_user.website

        user_info.promotion = (
            bool(await subscription_user.get_promotions())
            if isinstance(subscription_user, UserModel)
            else False
        )
        user_info.location = subscription_user.location
        user_info.website = subscription_user.website
        return user_info

    async def process_subscriptions(
        self,
        api_authed: ultima_scraper_api.auth_types,
        db_user: UserModel,
        performer_optimize: bool = False,
    ):
        if performer_optimize:
            api_subscriptions = await api_authed.get_subscriptions(filter_by="paid")
        else:
            api_subscriptions = await api_authed.get_subscriptions()
        with alive_bar(len(api_subscriptions)) as bar:
            for api_subscription in api_subscriptions:
                bar.title(
                    f"Processing Subscription: {api_subscription.user.username} ({api_subscription.user.id})"
                )
                await self.create_or_update_subscription(
                    api_subscription,
                    db_user,
                    performer_optimize=performer_optimize,
                )
                bar()

    async def process_content(
        self, db_user: UserModel, api_user: ultima_scraper_api.user_types
    ):
        if self.datascraper:
            session = self.get_session()
            content_manager = self.datascraper.resolve_content_manager(api_user)
            db_content_manager = self.resolve_content_manager(db_user)

            async def process_content_async(
                site_api: SiteAPI, db_user: UserModel, content: Any
            ):
                try:
                    await site_api.create_or_update_content(db_user, content)
                except Exception as _e:
                    breakpoint()
                    print(_e)

            async def process_media_async(
                site_api: SiteAPI, db_user: UserModel, media: "MediaMetadata"
            ):
                try:
                    await site_api.create_or_update_media(db_user, media)
                except Exception as _e:
                    breakpoint()
                    print(_e)

            async def process_filepath_async(
                site_api: SiteAPI, db_user: UserModel, media: "MediaMetadata"
            ):
                try:
                    await site_api.create_or_update_filepaths(db_user, media)
                except Exception as _e:
                    breakpoint()
                    print(_e)

            for _key, contents in content_manager.categorized.__dict__.items():
                _result = await asyncio.gather(
                    *[
                        process_content_async(self, db_user, content)
                        for content in contents.values()
                    ],
                    return_exceptions=True,
                )
                # await session.commit()
                for content in contents.values():
                    db_content = content.__db_content__
                    await db_content.awaitable_attrs.media
                    if (
                        isinstance(db_content, MassMessageModel)
                        and content.api_type == "Messages"
                    ):
                        stmt = delete(FilePathModel).where(
                            FilePathModel.mass_message_id == db_content.id
                        )

                        await session.execute(stmt)
                        pass
                        stmt = delete(ContentMediaAssoModel).where(
                            ContentMediaAssoModel.mass_message_id == db_content.id
                        )
                        await session.execute(stmt)
                        await session.delete(db_content)
                        await session.commit()

                _result2 = await asyncio.gather(
                    *[
                        process_media_async(self, db_user, media)
                        for content in contents.values()
                        for media in content.medias
                    ],
                    return_exceptions=True,
                )
                # await session.commit()
                _result2 = await asyncio.gather(
                    *[
                        process_filepath_async(self, db_user, media)
                        for content in contents.values()
                        for media in content.medias
                    ],
                    return_exceptions=True,
                )
                for _, content in contents.items():
                    await self.create_or_update_comment(content)
                session = self.get_session()
                await session.commit()
                for content in contents.values():
                    db_content = content.__db_content__
                    await db_content.awaitable_attrs.media
            size_sum = await db_content_manager.size_sum()
            if size_sum > 0:
                db_user.user_info.size = size_sum
            await session.commit()

    async def create_or_update_content(
        self, db_performer: UserModel, content: "ContentMetadata"
    ):
        api_performer = content.__soft__.get_author()
        receiver_id = content.receiver_id
        content_manager = db_performer.get_content_manager()
        found_db_content = await content_manager.find_content(
            content.content_id, content_type=content.api_type
        )
        if (
            isinstance(found_db_content, MessageModel)
            and found_db_content.receiver_id != api_performer.id
            and found_db_content.receiver_id != receiver_id
        ):
            raise Exception("Message receiver id mismatch")
        db_content = found_db_content or await content_manager.add_content(content)

        content.__db_content__ = db_content
        if isinstance(db_content, PostModel | MessageModel):
            db_content.update(content)
            if not db_content.paid:
                db_content.paid = True if content.paid else False

                assert self.datascraper
                # fix_missing_paid_content(
                #     db_content,
                #     content,
                #     datascraper=self.datascraper,
                #     api_performer=api_performer,
                # )
            db_content.price = content.price or 0
        if content.preview_media_ids:
            if "poll" not in content.preview_media_ids:
                pass
            pass
        if isinstance(db_content, MessageModel):
            db_content.verified = True
        db_content.created_at = content.created_at
        if not isinstance(db_content, StoryModel):
            db_content.media_count = content.__soft__.media_count

    async def create_or_update_media(self, db_user: UserModel, media: "MediaMetadata"):
        assert media.id
        db_content = None
        content_metadata = media.__content_metadata__
        if content_metadata:
            db_content = content_metadata.__db_content__
        media_manager = db_user.get_content_manager().media_manager
        found_media = media_manager.find_media(media.id)
        media_url = media.urls[0] if media.urls else None
        if not media_url and media.drm == False:
            return
        if not found_media:
            db_media = MediaModel(
                id=media.id,
                user_id=media.user_id,
                url=media_url,
                size=0,
                preview=media.preview,
                created_at=media.created_at,
            )
            try:
                db_user.medias.append(db_media)
            except Exception as _e:
                pass
            media_manager.add_media(db_media)
        else:
            db_media = found_media
            db_media.user_id = media.user_id
            db_media.created_at = media.created_at
        if not media.preview:
            db_media.url = media_url
        if db_content:
            found_media = await db_content.find_media(db_media.id)
            if not found_media:
                try:
                    db_content.media.append(db_media)
                except Exception as _e:
                    pass
        return db_media

    async def create_or_update_filepaths(
        self, db_user: UserModel, media: "MediaMetadata"
    ):
        assert media.id
        db_media = db_user.get_content_manager().media_manager.find_media(media.id)
        if not db_media:
            return
        db_content = None
        content_metadata = media.get_content_metadata()
        if content_metadata:
            content_info = (content_metadata.content_id, content_metadata.api_type)

            await db_media.awaitable_attrs.filepaths
            db_filepath = db_media.find_filepath(content_info)
            db_content = content_metadata.__db_content__
        else:
            db_filepath = db_media.find_filepath()
        if not db_filepath and media.filename:
            filepath = media.get_filepath()
            db_filepath = FilePathModel(
                filepath=filepath.as_posix(), preview=media.preview
            )
            if db_content:
                db_filepath.set_content(db_content)
            db_media.filepaths.append(db_filepath)
        return db_filepath

    async def create_or_update_comment(self, content: "ContentMetadata"):
        session = self.get_session()
        if isinstance(content.__soft__, post_types):
            if len(content.__soft__.comments) > 1:
                pass
            db_content = content.__db_content__
            assert db_content
            for comment in content.__soft__.comments:
                found_db_comment = await session.scalar(
                    select(CommentModel).where(CommentModel.id == comment.id)
                )
                if not found_db_comment:
                    giphy_id = None
                    reply_id = None
                    if isinstance(comment, OFCommentModel):
                        giphy_id = comment.giphy_id
                    else:
                        if comment.reply_id != comment.reply_root_id:
                            reply_id = comment.reply_id
                    db_comment = CommentModel(
                        id=comment.id,
                        post_id=db_content.id,
                        reply_id=reply_id,
                        user_id=db_content.user_id,
                        giphy_id=giphy_id,
                        text=comment.text,
                        likes_count=comment.likes_count,
                        created_at=comment.created_at,
                    )
                    session.add(db_comment)

    async def create_or_update_paid_content(
        self,
        api_authed: ultima_scraper_api.auth_types,
        db_user: UserModel,
        print_filter: list[str] = [],
    ):
        unique_user_ids: set[int] = set()
        paid_contents = await api_authed.get_paid_content()
        grouped_paid_contents = {}
        # Group paid_contents by author.id
        for paid_content in paid_contents:
            if isinstance(paid_content, dict):
                continue
            author_id = paid_content.get_author().id
            if author_id not in grouped_paid_contents:
                grouped_paid_contents[author_id] = []
            grouped_paid_contents[author_id].append(paid_content)
        with alive_bar(len(paid_contents)) as bar:
            for author_id, contents in grouped_paid_contents.items():
                supplier = await api_authed.get_user(author_id)
                local_user = await self.get_user(
                    supplier.id, load_content=True, load_media=True
                )
                await self.create_or_update_user(supplier, local_user)
                found_bought_content = await db_user.find_bought_content(supplier.id)
                if not found_bought_content:
                    bought_content = BoughtContentModel(supplier_id=supplier.id)
                    db_user.bought_contents.append(bought_content)
                unique_user_ids.add(supplier.id)
                for paid_content in contents:
                    if isinstance(paid_content, dict):
                        continue
                    bar.title(
                        f"Processing Paid Content: {db_user.username} ({db_user.id})"
                    )
                    if any(x in paid_content.text for x in print_filter):
                        urls: list[str] = []
                        for x in paid_content.media:
                            url: ParseResult | None = paid_content.url_picker(x)
                            if isinstance(url, ParseResult):
                                urls.append(url.geturl())
                        if urls:
                            print(urls, f"{paid_content.id}\n")
                    bar()
        await self.get_session().commit()
        for supplier_user_id in unique_user_ids:
            if self.aio_pika_wrapper:
                supplier_user = await self.get_user(supplier_user_id)

                message = create_notification(
                    "paid_content", api_authed.get_api().site_name, supplier_user
                )
                await self.aio_pika_wrapper.publish_notification(message)
            else:
                await self.create_or_update_notification(
                    "paid_content", supplier_user_id, api_authed
                )
        await self.get_session().commit()

    async def create_or_update_notification(
        self,
        category: str,
        supplier_user_id: int,
        api_authed: ultima_scraper_api.auth_types,
    ):

        db_supplier = await self.get_user(supplier_user_id, load_notifications=True)
        assert db_supplier
        found_notification = None

        for notification in db_supplier.notifications:
            if notification.category == category:
                if notification.authed_user_id == api_authed.id:
                    found_notification = notification
                    break
                elif notification.authed_user_id is None and not found_notification:
                    found_notification = notification

        if not found_notification:
            found_notification = NotificationModel(
                user_id=db_supplier.id,
                authed_user_id=api_authed.id,
                category=category,
            )
            db_supplier.notifications.append(found_notification)
        else:
            if not found_notification.authed_user:
                found_notification.sent_discord = False
                found_notification.sent_telegram = False
            found_notification.authed_user_id = api_authed.id
        return found_notification



from typing import TYPE_CHECKING

from fastapi import Request

from ultima_scraper_db.databases.rest_api import RestAPI

if TYPE_CHECKING:
    from ultima_scraper_collection.config import UltimaScraperCollectionConfig

    from ultima_scraper_db.databases.ultima_archive.database_api import ArchiveAPI


def get_ua_client(request: Request) -> "UAClient":
    return request.app


class UAClient(RestAPI):
    database_api: "ArchiveAPI"
    config: "UltimaScraperCollectionConfig"

    def __init__(self, database_api: "ArchiveAPI"):

        from ultima_scraper_db.databases.ultima_archive.api.app import routers

        super().__init__(
            root_path="./",
            proxy_headers=True,
        )
        self.include_routers(routers)
        self.database_api = database_api
        UAClient.database_api = self.database_api

    async def init(self, config: "UltimaScraperCollectionConfig"):
        from ultima_scraper_collection.managers.datascraper_manager.datascraper_manager import (
            DataScraperManager,
        )

        self.config = config
        self.datascraper_manager = DataScraperManager(
            self.database_api.server_manager, config
        )
        return self

    def select_site_api(self, site_name: str):
        datascraper = self.datascraper_manager.find_datascraper(site_name)
        assert datascraper is not None
        return datascraper.api



import argparse
import itertools
import random
from multiprocessing import Process
from pathlib import Path
from typing import TYPE_CHECKING, Any

import paramiko
import uvicorn
from alembic import command
from alembic.config import Config
from alembic.migration import MigrationContext
from alembic.operations.ops import MigrationScript
from alembic.runtime.environment import EnvironmentContext
from alembic.script.base import ScriptDirectory
from alembic.util import CommandError
from fastapi.middleware.cors import CORSMiddleware
from sqlalchemy import Connection, MetaData, inspect
from sqlalchemy.ext.asyncio import (
    AsyncEngine,
    AsyncSession,
    async_sessionmaker,
    create_async_engine,
)
from sshtunnel import SSHTunnelForwarder  # type: ignore

from ultima_scraper_db.databases.rest_api import RestAPI
from ultima_scraper_db.helpers import create_database, database_exists, test_connection

parser = argparse.ArgumentParser()
parser.add_argument("--dev", action="store_true", help="Enable dev mode")
args = parser.parse_args()

if TYPE_CHECKING:
    from ultima_scraper_db.databases.ultima_archive.api.client import UAClient


class Alembica:
    def __init__(
        self,
        database_path: str | Path,
        generate: bool = False,
        migrate: bool = False,
    ) -> None:
        self.database_path = Path(database_path)
        assert self.database_path.exists(), "Invalid database directory"
        self.migration_directory = self.database_path.joinpath("alembic")
        self.config = Config(self.database_path.joinpath("alembic.ini"))
        self.config.set_main_option(
            "script_location",
            self.migration_directory.as_posix(),
        )
        self.is_generate = generate
        self.is_migrate = migrate


class SSHConnection:
    def __init__(
        self,
        host: str,
        port: int,
        username: str,
        private_key: bytes,
        private_key_password: str | None = None,
    ) -> None:
        self.host = host
        self.port = port
        self.username = username
        self.private_key = private_key
        self.password = private_key_password


class Schema:
    def __init__(
        self,
        name: str,
        engine: AsyncEngine,
        session: AsyncSession,
        database: "Database",
    ) -> None:
        self.name = name
        self.engine = engine
        self.session = session
        self.sessionmaker = async_sessionmaker(engine, expire_on_commit=False)
        self.database = database


def show_schemas(conn: Connection):
    inspector = inspect(conn)
    schemas = inspector.get_schema_names()
    return schemas


class Database:
    def __init__(
        self,
        name: str,
        host: str,
        port: int,
        username: str,
        password: str,
        ssh: SSHTunnelForwarder | dict[str, Any] | None,
        metadata: MetaData | None = None,
        alembica: Alembica | None = None,
    ) -> None:
        self.name = name
        self.host = host
        self.port = port
        self.username = username
        self.password = password
        if isinstance(ssh, dict):
            ssh = self.handle_ssh(ssh, host, port)
        if ssh:
            ssh.start()
        self.ssh: SSHTunnelForwarder | None = ssh
        self.schemas: dict[str, Schema] = {}
        self.metadata = metadata
        self.alembica = alembica
        self._echo = True

    async def clone(self):
        database = Database(
            self.name,
            self.host,
            self.port,
            self.username,
            self.password,
            self.ssh,
            self.metadata,
            self.alembica,
        )
        return await database.init_db(self._echo)

    def handle_ssh(
        self, ssh_auth_info: dict[str, Any], local_host: str, local_port: int
    ):
        if ssh_auth_info["active"] and ssh_auth_info["host"]:
            private_key_filepath = ssh_auth_info["private_key_filepath"]
            ssh_private_key_password = ssh_auth_info["private_key_password"]
            private_key = (
                paramiko.RSAKey.from_private_key_file(
                    private_key_filepath, ssh_private_key_password
                ).key
                if private_key_filepath
                else None
            )
            random_port = random.randint(6000, 6999)
            ssh_obj = SSHTunnelForwarder(
                (ssh_auth_info["host"], ssh_auth_info["port"]),
                ssh_username=ssh_auth_info["username"],
                ssh_pkey=private_key,
                ssh_private_key_password=ssh_private_key_password,
                remote_bind_address=(local_host, local_port),
                local_bind_address=(local_host, random_port),
            )
            return ssh_obj
        else:
            return None

    def create_engine(
        self, connection_string: str, execution_options: dict[str, Any] = {}
    ) -> AsyncEngine:
        engine = create_async_engine(
            connection_string,
            echo=self._echo,
            pool_size=20,
            max_overflow=5,
            execution_options=execution_options,
        )
        return engine

    async def init_db(self, echo: bool = False):
        self._echo = echo
        if self.ssh:
            connection_string = f"postgresql+asyncpg://{self.username}:{self.password}@{self.ssh.local_bind_host}:{self.ssh.local_bind_port}/{self.name}"  # type: ignore
        else:
            connection_string = f"postgresql+asyncpg://{self.username}:{self.password}@{self.host}:{self.port}/{self.name}"
        # Use 'postgres' as the database for the test connection
        await test_connection(connection_string.replace(f"/{self.name}", "/postgres"))

        self.engine = self.create_engine(
            connection_string,
        )
        if not await database_exists(connection_string):
            confirm = input(
                f"Database '{self.name}' does not exist. Create it? (y/n): "
            )
            if confirm.lower() == "y":
                await self.setup()
            else:
                raise RuntimeError("Database setup aborted by user.")
        from sqlalchemy.schema import CreateSchema

        async with self.engine.connect() as conn:

            def create_schemas(session: Connection):
                if not self.metadata:
                    return
                for schema_name in self.metadata._schemas:  # type: ignore
                    session.execute(CreateSchema(schema_name, if_not_exists=True))
                session.commit()

            await conn.run_sync(create_schemas)
        await self.resolve_schemas()
        assert self.alembica
        if self.alembica.is_generate:
            await self.generate_migration()
            await self.resolve_schemas()
        if self.alembica.is_migrate:
            await self.run_migrations()
        await self.engine.dispose()
        return self

    async def resolve_schemas(self):
        connection_string = str(self.engine.url.render_as_string(hide_password=False))
        async with self.engine.connect() as conn:
            schema_strings: list[str] = await conn.run_sync(show_schemas)
        assert schema_strings
        for schema_string in schema_strings:
            engine = self.create_engine(
                connection_string,
                execution_options={
                    "application_name": schema_string,
                    "schema_translate_map": {None: schema_string},
                },
            )
            async_session = AsyncSession(engine, expire_on_commit=False)
            schema_obj = Schema(schema_string, engine, async_session, self)

            self.schemas[schema_obj.name] = schema_obj
        return self.schemas

    async def create(self):
        await create_database(self.engine.url.render_as_string(hide_password=False))

    def get_schema(self, name: str):
        return self.schemas[name.lower()]

    async def setup(self):
        await self.create()

    async def generate_migration(self):
        def get_current_revision(connection: Connection):
            context = MigrationContext.configure(connection)
            current_rev = context.get_current_revision()
            return current_rev

        def process_revision_directives(
            context: MigrationContext,
            revision: tuple[str, str],
            directives: list[Any],
        ):
            revisions: list[MigrationScript] = directives
            # Prevent actually generating a migration
            has_upgrades = list(
                itertools.chain.from_iterable(
                    op.as_diffs()
                    for script in revisions
                    for op in script.upgrade_ops_list
                )
            )
            if not has_upgrades:
                directives[:] = []

        def run_revision(
            connection: Connection,
            cfg: Config,
            autogenerate: bool,
            local_metadata: MetaData,
        ):
            cfg.attributes["connection"] = connection  # type:ignore
            revision = command.revision(
                alembic_cfg,
                autogenerate=autogenerate,
                process_revision_directives=process_revision_directives,
            )
            return revision

        async with self.engine.connect() as conn:
            current_rev = await conn.run_sync(get_current_revision)
            assert self.alembica
            alembic_cfg = self.alembica.config
            if not current_rev:
                try:
                    revision = await conn.run_sync(
                        run_revision, alembic_cfg, True, self.metadata
                    )
                    return revision
                except CommandError as _e:
                    await self.run_migrations()
                    pass
            else:
                revision = await conn.run_sync(
                    run_revision, alembic_cfg, True, self.metadata
                )
                return revision
            pass

    async def run_migrations(self) -> None:
        if not self.alembica:
            raise RuntimeError("Alembic configuration is missing.")

        alembic_cfg = self.alembica.config
        script = ScriptDirectory.from_config(alembic_cfg)

        def run_upgrade(connection: Connection, cfg: Config):
            cfg.attributes["connection"] = connection  # type:ignore
            command.upgrade(cfg, "head")
            return True

        try:
            head_rev = script.get_current_head()
            if head_rev:
                current_rev = script.get_revision(head_rev)
                with open(current_rev.path) as migration_file:
                    migration_content = migration_file.read().split(
                        "def downgrade() -> None:"
                    )[0]
                    if "drop_" in migration_content:
                        if (
                            input(
                                "Database migration contains drop statements. Continue? (y/n): "
                            ).lower()
                            != "y"
                        ):
                            raise RuntimeError("Migration aborted by user.")

            async with self.engine.connect() as conn:
                await conn.run_sync(run_upgrade, alembic_cfg)

        except Exception as e:
            print(f"Migration error: {e}")
            raise

    async def run_downgrade(self, version: str) -> None:
        def downgrade(connection: Connection, cfg: Config):
            cfg.attributes["connection"] = connection
            command.downgrade(alembic_cfg, version)

        async with self.engine.connect() as conn:
            assert self.alembica
            alembic_cfg = self.alembica.config
            await conn.run_sync(downgrade, alembic_cfg)


class DatabaseManager:
    def __init__(self) -> None:
        self.databases: dict[str, Database] = {}

    def create_database(
        self,
        name: str,
        host: str,
        port: int,
        username: str,
        password: str,
        ssh: SSHTunnelForwarder | dict[str, Any],
        metadata: MetaData | None = None,
        alembica: Alembica | None = None,
    ):
        database = Database(
            name, host, port, username, password, ssh, metadata, alembica
        )
        return database

    def add_database(self, database: Database):
        self.databases[database.name] = database
        return database

    def resolve_database(self, name: str):
        return self.databases[name]


def thread_function(fast_api: "UAClient", port: int):
    uvicorn.run(  # type: ignore
        fast_api,
        host="127.0.0.1" if not args.dev else "0.0.0.0",
        port=port,
        log_level="debug",
    )


class DatabaseAPI_:
    def __init__(self, database: Database) -> None:
        self.database = database

    def activate_api(self, fast_api: "RestAPI", port: int):

        from start import MIDDLEWARE

        for cls, opts in MIDDLEWARE:
            fast_api.add_middleware(cls, **opts)

        x = Process(
            target=thread_function,
            args=(
                fast_api,
                port,
            ),
            daemon=True,
        )
        x.start()
        self.server = x

    def find_schema(self, name: str):
        return self.database.schemas[name]



from typing import TYPE_CHECKING

from sqlalchemy import ScalarResult, Select, func, or_, select
from sqlalchemy.ext.asyncio import AsyncSession

from ultima_scraper_db.databases.ultima_archive.schemas.management import (
    HostModel,
    ServerModel,
    SiteModel,
)
from ultima_scraper_db.managers.database_manager import Schema

if TYPE_CHECKING:
    from ultima_scraper_collection import datascraper_types


class ManagementAPI:
    def __init__(
        self,
        schema: Schema,
        datascraper: "datascraper_types | None" = None,
    ) -> None:
        self.database = schema.database
        self.schema = schema
        self.datascraper = datascraper

    async def __aenter__(self):
        self._session: AsyncSession = self.schema.sessionmaker()
        return self

    async def __aexit__(self, exc_type: None, exc_value: None, traceback: None):
        await self._session.commit()
        await self._session.aclose()

    def resolve_session(self):
        assert self._session, "Session has not been set"
        return self._session

    def get_session(self):
        return self.resolve_session()

    async def get_sites(self):
        stmt = select(SiteModel)
        result = await self.get_session().execute(stmt)
        return result.scalars().all()

    async def get_server(self, server_id: int | None, server_name: str | None):
        stmt = select(ServerModel).where(
            or_(
                ServerModel.id == server_id,
                ServerModel.name == server_name,
            )
        )
        result = await self.get_session().scalar(stmt)
        return result

    async def get_hosts(self, host_id: int | None = None, host_name: str | None = None):
        stmt = select(HostModel)
        if host_id:
            stmt = stmt.where(HostModel.id == host_id)
        if host_name:
            stmt = stmt.where(HostModel.name == host_name)
        result = await self.get_session().scalars(stmt)
        return result.all()

    async def create_or_update_host(self, db_host: HostModel):
        stmt = (
            select(HostModel)
            .where(HostModel.name == db_host.name)
            .where(HostModel.identifier == db_host.identifier)
        )
        result = await self.get_session().scalar(stmt)
        if not result:
            self.get_session().add(db_host)
        return db_host



import re
from dataclasses import asdict, dataclass
from typing import Any, Generic, List, Literal, Optional, TypeVar

import ultima_scraper_api
from fastapi import APIRouter, Depends, HTTPException, Query, Request, Security
from fastapi.encoders import jsonable_encoder
from fastapi.routing import APIRouter
from pydantic import BaseModel, Field
from pydantic.generics import GenericModel
from sqlalchemy import nullslast, or_, orm, select
from sqlalchemy.sql import func, or_, select
from typing_extensions import Literal

from ultima_scraper_db.databases.ultima_archive.api.client import (
    UAClient,
    get_ua_client,
)
from ultima_scraper_db.databases.ultima_archive.api.routers.users import (
    AdvancedOptions,
    restricted,
)
from ultima_scraper_db.databases.ultima_archive.schemas.templates.site import (
    UserInfoModel,
)
from ultima_scraper_db.databases.ultima_archive.site_api import (
    MessageModel,
    PostModel,
    UserAliasModel,
    UserModel,
)
from ultima_scraper_db.helpers import extract_identifier_from_url

T = TypeVar("T")
router = APIRouter(
    prefix="/search",
    tags=["search"],
    responses={404: {"description": "Not found"}},
)


class PaginatedResponse(GenericModel, Generic[T]):
    total: int
    results: List[T]


class Filters:
    def __init__(
        self,
        site_name: str,
        q: str = Query(""),
        order_by: Optional[str] = Query(None),
        order_direction: str = Query("asc"),
        has_ppv: Optional[bool] = Query(None),
    ):
        self.site_name = site_name
        self.q = q
        self.order_by = order_by
        self.order_direction = order_direction
        self.has_ppv = has_ppv

    def as_dict(self, exclude_none: bool = False) -> dict[str, Any]:
        data = self.__dict__
        return (
            {k: v for k, v in data.items() if v is not None} if exclude_none else data
        )


@router.get("/users/{site_name}")
async def search_users(
    filters: Filters = Depends(),
    page: int = Query(1, ge=1),
    limit: int = Query(20, le=50),
    ua_client: UAClient = Depends(get_ua_client),
):
    site_name = filters.site_name
    q = filters.q
    order_by = filters.order_by or "id"
    order_direction = filters.order_direction
    has_ppv = filters.has_ppv

    database_api = ua_client.database_api
    site_api = ua_client.select_site_api(site_name)
    # Ensure valid order direction
    if order_direction.lower() not in ["asc", "desc"]:
        raise HTTPException(
            status_code=400, detail="Invalid order_direction, must be 'asc' or 'desc'."
        )
    offset = (page - 1) * limit  # calculate offset for pagination
    q = q.replace(" ", "")
    q = q.replace("@", "")
    q = extract_identifier_from_url(q)
    async with database_api.create_site_api(site_name) as site_db_api:

        # Create subqueries for both paid posts and messages
        paid_posts_subquery = (
            select(PostModel.user_id, func.count().label("posts_ppv_count"))
            .where(PostModel.paid.is_(True))
            .group_by(PostModel.user_id)
            .subquery()
        )

        paid_messages_subquery = (
            select(MessageModel.user_id, func.count().label("messages_ppv_count"))
            .where(MessageModel.paid.is_(True))
            .group_by(MessageModel.user_id)
            .subquery()
        )
        stmt = (
            select(
                UserModel,
                (
                    func.coalesce(paid_posts_subquery.c.posts_ppv_count, 0)
                    + func.coalesce(paid_messages_subquery.c.messages_ppv_count, 0)
                ).label("ppv_count"),
            )
            .outerjoin(
                paid_posts_subquery, UserModel.id == paid_posts_subquery.c.user_id
            )
            .outerjoin(
                paid_messages_subquery, UserModel.id == paid_messages_subquery.c.user_id
            )
            .outerjoin(UserAliasModel)
            .where(
                or_(
                    UserModel.username.ilike(f"%{q}%"),
                    UserAliasModel.username.ilike(f"%{q}%"),
                )
            )
        )
        if has_ppv:
            # Only include users who have at least one paid post or paid message
            stmt = stmt.where(
                or_(
                    paid_posts_subquery.c.posts_ppv_count.isnot(None),
                    paid_messages_subquery.c.messages_ppv_count.isnot(None),
                )
            )

        if order_by in ["downloaded_at", "size"]:
            field = (
                UserModel.downloaded_at
                if order_by == "downloaded_at"
                else UserInfoModel.size
            )
            if order_direction.lower() == "desc":
                stmt = stmt.order_by(nullslast(field.desc()), UserModel.id)
            else:
                stmt = stmt.order_by(field, UserModel.id)
        else:
            stmt = stmt.order_by(UserModel.id)

        total_stmt = select(func.count()).select_from(stmt.subquery())
        total = await site_db_api.get_session().scalar(total_stmt)
        stmt = (
            stmt.offset(offset)
            .limit(limit)
            .options(*restricted)
            .options(orm.selectinload(UserModel.user_info))
        )
        results = await site_db_api.get_session().execute(stmt)
        final_users: list[dict[str, Any]] = []
        accurate_username = False
        for user, ppv_count in results:
            temp_user = jsonable_encoder(user)
            temp_user["user_info"]["ppv_count"] = ppv_count
            if q == user.username:
                accurate_username = True
            final_users.append(temp_user)
        if not final_users or not accurate_username:
            async with site_api.login_context(guest=True) as authed:
                if authed:
                    site_user = await authed.get_user(q)
                    if site_user:
                        db_user = await site_db_api.get_user(
                            site_user.id,
                            extra_options=restricted,
                        )
                        user = await site_db_api.create_or_update_user(
                            site_user, db_user, performer_optimize=True
                        )
                        user.content_manager = None
                        await user.awaitable_attrs.aliases
                        await user.awaitable_attrs.user_info
                        await user.awaitable_attrs.remote_urls
                        results = await site_db_api.get_session().scalars(stmt)
                        final_users = [
                            jsonable_encoder(user, exclude={"user_info": "user"})
                            for user in results.all()
                        ]

        return PaginatedResponse[dict[str, Any]](
            total=total, results=jsonable_encoder(final_users)
        )


@router.get(
    "/{site_name}/content",
    response_model=PaginatedResponse[dict[str, Any]],
)
async def search_content(
    site_name: str,
    category: Literal["posts", "messages"],
    q: str = "",
    user_id: int | None = None,
    ppv: bool | None = None,
    page: int = Query(1, ge=1),
    limit: int = Query(100, le=1000),
    ua_client: UAClient = Depends(get_ua_client),
):
    database_api = ua_client.database_api
    offset = (page - 1) * limit

    async with database_api.create_site_api(site_name) as site_db_api:
        session = site_db_api.get_session()

        if category == "posts":
            model = PostModel

        elif category == "messages":
            model = MessageModel

        else:
            raise HTTPException(status_code=400, detail="Invalid content type")

        stmt = select(model)
        if user_id:
            stmt = stmt.where(model.user_id == user_id)
        if q:
            stmt = stmt.where(model.text.ilike(f"%{q}%"))
        if ppv is not None:
            stmt = stmt.where(model.price > 0 if ppv else model.price == 0)

        total_stmt = select(func.count()).select_from(stmt.subquery())
        total = await session.scalar(total_stmt)

        stmt = (
            stmt.order_by(nullslast(model.created_at))
            .offset(offset)
            .limit(limit)
            .options(orm.noload(model.media))
        )

        results = await session.scalars(stmt)
        posts = results.unique().all()

        return PaginatedResponse[dict[str, Any]](
            total=total, results=jsonable_encoder(posts)
        )
